
"""
cells - Module to decribe symplectic cells

AUTHOR:

- JENS PUTZKA (2009-09-10): initial version

DATE:
- 2011-03-22

VERSION:
- 1.61 -2012-05-13
- 1.6 - 2011-03-22
- 1.5 - 2010-10-06
- 1.1 - 2010-04-27
- 1.0 - 2009-09-10


"""
#*****************************************************************************
#       Copyright (C) 2009 - JENS PUTZKA <putzka@mpim-bonn.mpg.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

print "\n=========================================================";
print '|    Importing Cells Package  (c) 2008-2011             |';
print '|                                                       |'
print '|    by Jens Putzka MPI for Mathematics Bonn            |'
print "=========================================================";
#===============================================================================
# Imports
#===============================================================================

from sage.rings.all import ZZ,QQ

from config import BASEPATH
from sage.misc.latex import latex

from sage.matrix.all import is_Matrix, diagonal_matrix, matrix, identity_matrix
from sage.structure.sage_object import SageObject
from sage.combinat.all import tuples, arrangements
from sage.modules.free_module_element import vector
from sage.groups.all import MatrixGroup, is_MatrixGroupElement,is_MatrixGroup
from sage.parallel.all import parallel
from sage.interfaces.all import gap
from io import saveobj2file
from io import loadobj 


from symplectic import is_symplectic, is_GSp

#===============================================================================
# PART I: Functions and Constants
#===============================================================================


#===============================================================================
# sgn4 ()
#===============================================================================

sgn4 = tuples([-1,1],4)

#=====================================================
# is_Cell()
#=====================================================

def is_Cell(c):
    return isinstance (c,Cell);


#=====================================================
# getstandardform()
#=====================================================

def getstandardform (matr):
    """"
    Converts matrix to matrix in normed form
    The normation is given by:
    * the first non zero element in each column is positive
    * the columns are ordered first by numbers of zeroes, then by the size 1st elements, the the size of 2nd elements and so on in all columns.  
    
    """
        
    if not is_Matrix(matr):
        raise TypeError, "matr has to be a matrix.";
    _matr_cols  = matr.columns();
    _normed_col = [];
    _n          = matr.nrows();
    _m          = matr.ncols();
    ########################
    #Normalizing columns
    ########################
    for col in _matr_cols:
        for i in range(_n):
            if col[i]<0:
                _normed_col.append((-vector(col)).list());
                break;
            elif col[i]==0:
                continue;
            else:
                _normed_col.append(col.list());
                break;
    ########################
    #Sorting
    ########################
    _sort_key = lambda x : tuple ([x.count(0)]+[x[i] for i in range(_n)]);
    _sorted_list = sorted(_normed_col, key = _sort_key,reverse=True);
    _out_list=[];
    for vec in _sorted_list:
        _out_list+=vec;
    _new_matr=matrix(_m,_out_list).transpose();    
    return _new_matr;

 
#===============================================================================
# computesymplecticmap2standard ()
#===============================================================================


def computesymplecticmap2standard (cell, symplectic_type = 0):
    """
    Computes a map to a standard cell. It is recommanded to use getmap2standard
    which tries to obtain a result faster, since it tries to get it from known results.
    """
    
    
    _n_rows = cell.nrows();

    
    sgn_mat_list = [diagonal_matrix(d) for d in tuples([-1,1],_n_rows)]
    _sgn = sgn_mat_list[:]
    sgn_s = [sgn_mat_list.pop(sgn_mat_list.index(m)) for m in _sgn if is_symplectic(m, symplectic_type = symplectic_type)]
    sgn_ns = sgn_mat_list[:]
    
    print "Computing list of matrices for possible surrounding the top dimensional cell\n"
    mat = cell.matrix()
    _tmp = [mat.matrix_from_columns(ar) for ar in arrangements(range(mat.ncols()),_n_rows)]
    _tmp = [m for m in _tmp if m.rank()==_n_rows]
    print "   --> split list in symplectic and non symplectic";
    _tmp_0 = _tmp[:]
    list_s =[_tmp.pop(_tmp.index(m)) for m in _tmp_0 if is_symplectic(m, symplectic_type = symplectic_type)];
    list_n = _tmp[:];
    del _tmp;   
    print "    --> done";
        
    for m in list_s:
        for d in sgn_s:
            __mat = (d * m).inverse()
            new_cell = cell.apply(__mat)
            if new_cell in StandardElementList:
                return __mat
            elif new_cell in CellList:
                __mat2 = CellList[CellList.index(new_cell)].map2standard()
                if __mat2 != None:
                    return __mat2*__mat
                else:
                    continue
            else:
                continue
    for m in list_s:
        for d in sgn_s:
            ___mat = (d * m)
            if not is_symplectice(___mat):
                continue
            __mat=__mat.inverse()
            new_cell = cell.apply(__mat)
            if new_cell in StandardElementList:
                return __mat
            elif new_cell in CellList:
                __mat2 = CellList[CellList.index(new_cell)].map2standard()
                if __mat2 != None:
                    return __mat2*__mat
                else:
                    continue
            else:
                continue                
    raise RuntimeError # Should never occur
     
    
    
#===============================================================================
# getsymplecticmap2standard ()
#===============================================================================

def getmap2standard (cellin):
    """
    getsymplecticmap2standard
    
    INPUT:
    * Cell
    
    RETURN:
    * Symplectic matrix which maps the given cell to its standard cell
    
    ALGORITHM:
    1. Check wether the stored matrixwhich is by construction in GSp is already symplectic.
    2. If return that matrix if not try to change elements by elements of stabilisers on both sides.
    3. If that does not work one starts the computation.
    """
    
    
    if not is_Cell(cellin):
        raise TypeError;
    matr = cellin.map2standard()
    if matr != None:
        if is_symplectic(matr):
            return matr
        elif is_Matrix(matr):
            if is_GSp(matr):
                for g in cellin.stabilizerGSp():
                    gnew =matr*g
                    if is_symplectic(gnew):
                        setattr(cellin, '_closure', gnew)
                        return gnew
                computesymplecticmap2standard (cellin)
            else:
                computesymplecticmap2standard (cellin)
        else:
            raise TypeError
    else:
        computesymplecticmap2standard (cellin)

    return gnew



#=====================================================
# getclosure()
#=====================================================

def getclosure (cellin, basedir = BASEPATH, dir = 'output/cells/closures'):
    """
    RETURN:
    List of cells which lie in the boundary of the cell grouped by.
    optional parameter where to search path =
    """
    
    if not cellin.is_in_closure_of(Cell(identity_matrix(4))):
        raise NotImplementeError , "At the moment it is not possible to compute the closure of cells which do not ly in the closure of the StandardCell RedSquare. Sorry."
        
    if cellin._closure != None:   
        return cellin._closure;
    print 'Start: Computing Closure of %s' %cellin;


    
    #######################################################################
    # Try to load closure for this cell precomputed to be found in output/cells/closures
    #######################################################################
    
    try:
        print '   * Try loading closure from file'
        _loc_cls = loadobj(cellin.name()+'closure' , basedir = basedir, dir = dir)
        print '   * File loaded'
        _out =  [c for c in _loc_cls if c.is_in_closure_of(cellin)]
        print '   * Set attributes\n'
        setattr(cellin, '_closure', _out)


        print '---> Done\n'
        return _out        
    except:
        print 'Closure is not stored for %s\n\n' %cellin  
       
    #######################################################################    
    # Try to load file for topdimesnional cell and deduce closure from that
    #######################################################################    
    
    try:
        print '   * Try loading closure of RedSquare from file';
        _loc_cls = loadobj('RedSquare0closure', basedir = basedir, dir = dir)
        print '   * File loaded';
        print '   * Taking closure of %s in the closure of RedSquare0' %cellin;
        _out =  [c for c in _loc_cls if c.is_in_closure_of(cellin)]
        print '   * Set attributes\n';
        setattr(cellin, '_closure', _out)
        
        print '   * Save closure for later use'
        saveobj2file (_out, cellin.name()+'closure' , basedir = basedir, dir = dir)
        print '---> Done\n'
        return _out
    except:
        print '   * File not found!\n\n'
        print '   Start: New Computation of closure of %s' %cellin;

    
    ####################################################################### 
    # Start with the real computation
    ####################################################################### 
           
    _n_rows = cellin.nrows();

    #############################################
    # Define sign matrices
    #############################################
    
    _sgn = [];
    for sgn in tuples([-1,1],_n_rows):
        """
        +- define the same cell therefore only one of both has to be taken. 
        Reduces size to half of the original size.
        """
        d=diagonal_matrix(sgn);

        if d not in _sgn:
            if -d not in _sgn:
                _sgn.append(d);
    #############################################
    # Filter sign matrices to Sp or GSp
    #############################################   
    print "Generate signs";

    _sgn_mat_symp     = [m for m in _sgn if is_symplectic(m)];
    _sgn_mat_non_symp = [m for m in _sgn if not is_symplectic(m)];
    del _sgn;   

    #############################################
    # Run over cells which can not excluded 
    # by their dimension take every 4-tuple
    # multiply with a sign computed before.
    # Compute the image of the cell under that
    # map.
    #############################################
    print "Start run over elements" 
    _out=[cellin]; #geaendert von []
    for cell in StandardElementList:
        
        if cell.dim() >= cellin.dim():
            continue
        
        print "* Elements of type %s" %cell.cell_type();
            
        _i = getindex(cell);  #=0
        #############################################
        # Filter to those matrices from columns
        # which are symplectic and those which are not
        # since 
        #############################################
        print "   - Filter matrices";
        cell_list  = [ cell._matrix.matrix_from_columns(arr) for arr in arrangements(range(cell.ncols()),_n_rows)];#arrangments().list()
        cell_list  = [ m for m in cell_list if m.det().abs()==1];
        cell_list0 = cell_list[:];
                        
        
        cell_list_s =[cell_list.pop(cell_list.index(m)) for m in cell_list0 if is_symplectic(m)];
        cell_list_n = cell_list[:];
        del cell_list;


        _cell_matrix_list =[];
        print "   - Check for symplectic matrices: Part 1 (symp.)";    
        for m in cell_list_s:
            _new_cell = m.inverse()*cell._matrix;
            for d in _sgn_mat_symp:
                _new_cell_d = getstandardform(d*_new_cell);
                    
                if all([column in _new_cell_d.columns() for column in cellin._matrix.columns()]):
                    if _new_cell_d not in _cell_matrix_list :
                        _cell_matrix_list.append(_new_cell_d); 
                        __out_map2std = m*d;

                        _NC = Cell (_new_cell_d, name =  cell._cell_type+str(_i), latex_name = cell._latex_+'_{%s}'%_i, cell_rank = cell._rank, cell_type = cell._cell_type, transv = cell._transv, bd_type = cell._bd_type, map2standard = __out_map2std);
                        _out.append(_NC)
                        add2knowncells(_NC)
                        _i += 1;
        print "   - Check for symplectic matrices: Part 2 (non symp.)";              
        for m in cell_list_n:
            try:
                _new_cell = ((m.inverse()).change_ring(ZZ))*cell._matrix    
            except:
                _new_cell = m.inverse()*cell._matrix
            for d in _sgn_mat_non_symp: 
                _new_cell_d = getstandardform(d*_new_cell)

                __out_map2std = m*d;
                if is_GSp(__out_map2std):
                    if all([column in _new_cell_d.columns() for column in cellin._matrix.columns()]):
                        if _new_cell_d not in _cell_matrix_list :
                            _cell_matrix_list.append(_new_cell_d); 
                            _NC = Cell (_new_cell_d, name =  cell._cell_type+str(_i), latex_name = cell._latex_+'_{%s}'%_i, cell_rank = cell._rank, cell_type = cell._cell_type, transv = cell._transv, bd_type = cell._bd_type, map2standard = __out_map2std);
                            _out.append(_NC);
                            add2knowncells(_NC)
                            _i += 1;       
        #print '4';                
        print '-->Done';                      
            
    
    ####################################################################### 
    # Postprocessing    
    ####################################################################### 
    print "\n\nPOSTPROCESSING:\n\n";        
    print ' * Set attributes\n';
    setattr(cellin, '_closure', _out); 
    print ' * Save closure for later use'

    saveobj2file (_out, cellin.name()+'closure' , basedir = basedir , dir = dir) 
    print '-->Done'; 
    return cellin._closure;



#===============================================================================
# getstabilizer ()
#===============================================================================


def getstabilizer (cellin, basedir = BASEPATH, dir = 'output/cells/stabilizers'):
    """
    Return Stabilzer of a given cell as MatrixGroup which is a subgroup of the symplectic group
    """
  
    if not is_Cell(cellin):
        raise TypeError, "%s has to be a cell" %cellin

    if cellin._stab != None:
        return cellin._stab;
           
    if not cellin.is_in_closure_of(Cell(identity_matrix(4))):
        raise NotImplementeError , "At the moment it is not ossible to compute the stabilizer of cells which do not ly in the closure of the StandardCell RedSquare. Sorry."

    ####################################################################### 
    #Try Loading Elements
    ####################################################################### 

    try:
        print '   * Try loading closure from file'
        _grp = loadobj(cellin.name()+'stabilizer' , basedir = basedir, dir = dir)
        print '   * File loaded'
        print '   * Set attributes\n'
        setattr(cellin, '_stab', _grp)
        print '---> Done\n'
        return _grp        
    except:
        print '-> Stabilizer is not stored for %s\n\n' %cellin  

    #######################################################################     
    # Produce elements  
    #######################################################################   
    print "Start computing stabilizer.\n\n"
    _stab_list = [];
    _sgn_list  = tuples([-1,1],cellin.nrows());
    _perm_list = arrangements(range(cellin.ncols()),cellin.nrows());

    for perm in _perm_list:
        #ZZ eingefuegt
        try:
            _perm_mat = cellin.matrix(R=ZZ).matrix_from_columns(perm);
        except:
            _perm_mat = cellin.matrix().matrix_from_columns(perm);

        for sgn in _sgn_list:
            _perm_mat_sgn = _perm_mat*diagonal_matrix(sgn);

            if is_symplectic(_perm_mat_sgn):
                if cellin.matrix()==getstandardform(_perm_mat_sgn * cellin.matrix()):
                    if _perm_mat_sgn not in _stab_list:
                        _stab_list.append(_perm_mat_sgn);
    print "---> Done\n"
    ####################################################################### 
    # Postprocessing :    Make group from group Elements
    #######################################################################     
    _grp =  MatrixGroup(_stab_list);
    _grp.rename('stabilizer of cell %s of order %s' % (cellin, _grp.order()));
    cellin.set_stabilizer(_grp);  
    
    print '   * Save stabilizer for later use'

    saveobj2file (_grp, cellin.name()+'stabilizer' , basedir = basedir , dir = dir) 
    print '    -->Done'; 
    
    return _grp;            


#===============================================================================
# getstabilizerGSp ()
#===============================================================================


def getstabilizerGSp (cellin, basedir = BASEPATH,  dir = 'output/cells/stabilizers'):
    """
    Return Stabilzer of a given cell as MatrixGroup
    """
  
    if not is_Cell(cellin):
        raise TypeError, "%s has to be a cell" %cellin

    if cellin._stab != None:
        return cellin._stab;
           
    if not cellin.is_in_closure_of(Cell(identity_matrix(4))):
        raise NotImplementeError , "At the moment it is not ossible to compute the stabilizer of cells which do not ly in the closure of the StandardCell RedSquare. Sorry."

    ####################################################################### 
    #Try Lopading Elements
    ####################################################################### 

    try:
        print '   * Try loading closure from file'
        _grp = loadobj(cellin.name()+'stabilizerGSp' , basedir = basedir, dir = dir)
        print '   * File loaded'
        print '   * Set attributes\n'
        setattr(cellin, '_stab', _grp)
        print '---> Done\n'
        return _grp        
    except:
        print '-> Stabilizer is not stored for %s\n\n' %cellin  

    #######################################################################     
    # Produce elements  
    #######################################################################   
    print "Start computing stabilizer.\n\n"
    _stab_list = [];
    _sgn_list  = tuples([-1,1],cellin.nrows());
    _perm_list = arrangements(range(cellin.ncols()),cellin.nrows());

    for perm in _perm_list:
        try:
            _perm_mat = cellin.matrix(R=ZZ).matrix_from_columns(perm);
        except:
            _perm_mat = cellin.matrix().matrix_from_columns(perm);

        for sgn in _sgn_list:
            _perm_mat_sgn = _perm_mat*diagonal_matrix(sgn);
            if is_GSp(_perm_mat_sgn):
                if cellin.matrix()==getstandardform(_perm_mat_sgn * cellin.matrix()):
                    if _perm_mat_sgn not in _stab_list:
                        _stab_list.append(_perm_mat_sgn);
    print "---> Done\n"
    ####################################################################### 
    # Postprocessing :    Make group from group Elements
    #######################################################################     
    _grp =  MatrixGroup(_stab_list);
    _grp.rename('stabilizer of cell %s of order %s' % (cellin, _grp.order()));
    cellin.set_stabilizer(_grp);  
    
    print '   * Save stabilizer for later use'
    
    saveobj2file (_grp, cellin.name()+'stabilizerGSp' , basedir = basedir , dir = dir) 
    print '    -->Done'; 
    
    return _grp;           


#===============================================================================
# listofsymplecticmatricesfromcolumns ()
#===============================================================================

def listofsymplecticmatricesfromcolumns (mat,symplectic_type=0):
    """
    Returns a list of symplectic matrice which are build out of column vectors of mat (up to sign)
    """

    _n_rows = mat.nrows();
    
    sgn_mat_list = [diagonal_matrix(d) for d in tuples([-1,1],_n_rows)]
    _sgn = sgn_mat_list[:]
    sgn_s = [sgn_mat_list.pop(sgn_mat_list.index(m)) for m in _sgn if is_symplectic(m, symplectic_type = symplectic_type)]
    sgn_ns = sgn_mat_list[:]
    
    print "Computing list of matrices for possible surrounding the top dimensional cell\n"

    _tmp = [mat.matrix_from_columns(ar) for ar in arrangements(range(mat.ncols()),_n_rows)]
    print "   * All %sx%s matrices from columns %10s" % (_n_rows,_n_rows,len(_tmp))
    _tmp = [m for m in _tmp if m.rank()==_n_rows]
    print "   * All %sx%s matrices of full rank from columns %10s" % (_n_rows,_n_rows,len(_tmp))
    print "   --> split list in symplectic and non symplectic";
    _tmp_0 = _tmp[:]
    list_s =[_tmp.pop(_tmp.index(m)) for m in _tmp_0 if is_symplectic(m, symplectic_type = symplectic_type)];
    list_n = _tmp[:];
    del _tmp;   
    print "    --> done";
    print "Add symplectic products to list"
    out = [d*m for m in list_s for d in sgn_s]
    print "    --> done\n";
    print "Start to check symplecticity for the remaining cases.\n";
    for d in sgn_ns:
        for m in list_n:
            md= d*m
            if is_symplectic(md, symplectic_type = symplectic_type):
                out.append(md)
    
    print " Done\n"
    print "   * Found number of symplectic  %sx%s matrices from columns is %10s\n" % (_n_rows,_n_rows,len(out))

    return out







#===============================================================================
# listofGSpmatricesfromcolumns ()
#===============================================================================


def listofGSpmatricesfromcolumns (mat, symplectic_type=0):
    """
    Returns a list of symplectic matrice which are build out of column vectors of mat (up to sign)
    """

    _n_rows = mat.nrows();
    
    sgn_mat_list = [diagonal_matrix(d) for d in tuples([-1,1],_n_rows)]
    _sgn = sgn_mat_list[:]
    sgn_s = [sgn_mat_list.pop(sgn_mat_list.index(m)) for m in _sgn if is_symplectic(m, symplectic_type = symplectic_type)]
    sgn_ns = sgn_mat_list[:]
    
    print "Computing list of matrices for possible surrounding the top dimensional cell\n"
    
    _tmp = [mat.matrix_from_columns(ar) for ar in arrangements(range(mat.ncols()),_n_rows)]
    print "   * All %sx%s matrices from columns %10s" % (_n_rows,_n_rows,len(_tmp))
    _tmp = [m for m in _tmp if m.rank()==_n_rows]
    print "   * All %sx%s matrices of full rank from columns %10s" % (_n_rows,_n_rows,len(_tmp))
    print "   --> split list in symplectic and non symplectic";
    _tmp_0 = _tmp[:]
    list_s =[_tmp.pop(_tmp.index(m)) for m in _tmp_0 if is_GSp(m, symplectic_type = symplectic_type)];
    list_n = _tmp[:];
    del _tmp;   
    print "    --> done";
    print "Add symplectic products to list"
    out = [d*m for m in list_s for d in sgn_s]
    print "    --> done\n";
    print "Start to check symplecticity for the remaining cases.\n";
    for d in sgn_ns:
        for m in list_n:
            md = d*m
            if is_GSp(md, symplectic_type = symplectic_type):
                out.append(md)
    
    print " Done\n"
    print "   * Found number of symplectic  %sx%s matrices from columns is %10s\n" % (_n_rows,_n_rows,len(out))

    return out

#===============================================================================
# list_from_columns ()
#===============================================================================

def list_from_columns (mat, symplectic_type=0):
    _n_rows = mat.nrows();
    
    sgn_mat_list = [diagonal_matrix(d) for d in tuples([-1,1],_n_rows)]
    _tmp = [mat.matrix_from_columns(ar) for ar in arrangements(range(mat.ncols()),_n_rows)]
    out = []
    for d in sgn_mat_list:
        for m in _tmp:
            dm = d*m
            if is_GSp(dm):
                out.append(dm)
    print len(out)
    return out
    

#===============================================================================
# matrix_list_2_cellmatrix_list ()
#===============================================================================

def matrix_list_2_cellmatrix_list (lst):
     
    _tmp = [getstandardform(t) for t in lst]
    
    _out = [];
    for t in _tmp:
        if t not in _out:
            _out.append(t)     

    return _out
    
#===============================================================================
# getneighbours ()
#===============================================================================


def getneighbours(cellin, type=None, basedir = BASEPATH, dir = 'output/cells/neighbours'):
    """
    Compute the neighbors of cellin of type type or all if type=None (default)
    
    If type is none it returns all cells of higher dimension lying 
    around cellin otherwise only thos of type type.
    """
    
    
    if not is_Cell(cellin):
        raise ValueError, "%s has to be a cell" %cellin
        
    
    
    ####################################################################### 
    #Try Lopading Elements
    ####################################################################### 
    if cellin._nb!=None:
        if type == None:
            return cellin._nb
        else:
            out = [c for c in cellin._nb if c.cell_type()==type]
            return out
        
    try:
        print '   * Try loading neighbours from file'
        _nb = loadobj(cellin.name()+'neighbours-'+'%s'%type , basedir = basedir, dir = dir)
        print '   * File loaded'
        if type==None:
            print '   * Set attributes\n'
            setattr(cellin, '_nb', _nb)
            print '---> Done\n'
        return _nb        
    except:
        print '-> Neighbours are not stored for %s\n\n' %cellin  
    
    """    
    try:
    Loading from cell of the same type and translate
    """    

    #######################################################################     
    # Produce elements  
    #######################################################################   
    if type != None and cellin.cell_type()==type:
        return []
    elif type == None:
        print "======================================================"
        print "|   Compute List of symplectic matrices from columns "
        sympMats = [m.inverse() for m in listofsymplecticmatricesfromcolumns(cellin.matrix())]
        print "|   --->done"
        print "======================================================\n"   
        out =[]   
        print "====================================================="
        print "|    * Start loop over cells"          
        for c in CellList:
            #print "*";
            if cellin.dim() >= c.dim():
                continue    
            
            for m in sympMats:
                _m = c.apply(m)
                if cellin.is_in_closure_of(_m):
                    if _m not in out:
                        out.append(_m)
        print "\n";
    else:# different types
        stdC = [c for c in StandardElementList if c.cell_type()== type][0]
        d = stdC.dim();
        print "d=%s" %d;
        if cellin.dim() >= d:
            print "E"
            out = []
        print "====================================================="
        print "| Compute List of symplectic matrices from columns"
        sympMats = listofsymplecticmatricesfromcolumns(cellin.matrix())
        print "|   --->done"
        print "====================================================="

        print "====================================================="
        print "|    * Start loop over cells"        

        out = [] 
        for m in sympMats:
            _m = stdC.apply(m)
            if cellin.is_in_closure_of(_m):
                if _m not in out:
                    out.append(_m)
        print "|    --> Done";    
        print "====================================================="           

 
    #######################################################################     
    # Write elements  
    #######################################################################  
    
    print '   * Set attributes\n'
    if type== None:
        print '   * Set attributes\n'
        setattr(cellin, '_nb', out)
     
    print '   * Save neighbours for later use'
    saveobj2file (out, cellin.name()+'neighbours-'+'%s'%type , basedir = basedir, dir = dir)
    print '---> Done\n'

    return out

#===============================================================================
# decomposeintoorbits ()
#===============================================================================

def decomposeintoorbits(SetOfCells, Grp):
    """
    INPUT::
        SetOfCells = a set of cells
        Grp        = a matrix group or a set of matrices
    OUPTPUT::
        A set of representatives for SetOfCells modulo Grp
    
    """
    C   = SetOfCells[:]
    out = [] 
    cnt=1
    while C!=[]:
        orb = [] 
        c0= C.pop(0)
        for g in Grp:
            gc0= c0*g
            
            if gc0 not in SetOfCells:
                print "Caution: Set of cells was not closed under the group action. \nTherefore new allements where added."
                if gc0 not in orb:
                    orb.append(gc0)
            else:
                if gc0 not in orb:
                    orb.append([c for c in SetOfCells if c==gc0][0])
                if gc0 in C:
                    C.remove(gc0)
        out.append(orb)
        cnt=cnt+1
    
    return out


#===============================================================================
# PART II: Classes
#===============================================================================

#===============================================================================
# Cell()
#===============================================================================

class Cell (SageObject):
    def __init__ (self, matrix, name = "Cell", latex_name = "Cell", cell_rank = None, cell_type = None, transv = None, bd_type = False, list_of_perm = None ,map2standard = None , closure = None, stabilizer = None, neighbours= None): 
        self._matrix     = getstandardform(matrix);
        self._rank       = cell_rank;
        self._name       = name;
        self._latex_      = latex_name;
        self._cell_type  = cell_type;
        self._bd_type    = bd_type;
        self._transv     = transv;
        self._map2standard = map2standard;    
        self._stab       = stabilizer;
        self._closure    = closure;
        self._nb         = neighbours
            
    #=====================================================
    # Ways to represent the cell
    #=====================================================
        
    def _repr_ (self):
        return self._name;    

    def name (self):
        return self._name;  
    
    def _latex_ (self):
        return self._latex_;      

    def latex (self):
        return self._latex_;
        
    def _matrix_ (self, R = None):
        if R is None:
            return self._matrix;
        else:
            return self._matrix.change_ring(R);

    def matrix (self, R= None) :
        if R is None:
            return self._matrix;
        else:
            return self._matrix.change_ring(R);     

    def matrix_latex (self):
        latex.matrix_delimiters("[", "]");
        out= latex(self.matrix());
        latex.matrix_delimiters("(", ")");
        return out

    def map2standard(self):
        """
        RETURN:
        
        Matrix such that
        
        cell.map2standard()*cell 
        
        is one of the standard cells. 
        
        REMARK:
        At the moment if no such map is given at initialization of the cell 
        it has value None and this function returns a NotImplementedError.
        """
        
        if self._map2standard != None:
            return self._map2standard
        else:
            getmap2standard(self)

    #=====================================================
    # Change names etc.
    #=====================================================

    def rename (self, new_name):
        if not isinstance(new_name, str):
            raise TypeError, "new name has to be a string.";
        self._name = new_name;    

    def rename_latex (self, new_name):
        if not isinstance(new_name, str):
            raise TypeError, "new name has to be a string.";
        self._latex_ = new_name;   

    def set_map2standard (self, matrix):
        #One could include some some checks for validity of the input;
        self._map2standard = matrix;



    def set_closure(self, lst):
        #One could include some some checks for validity of the input;
        self._closure = lst;
        print self._closure;

    def set_stabilizer (self, grp):
        #One could include some some checks for validity of the input;
        self._stab = grp; 
    
    #=====================================================
    # Information on the Cell
    #=====================================================

    def cell_type (self):
        return self._cell_type;  

    def is_boundary_type (self) :
        return self._bd_type;

    def is_transverse (self) :
        return self._transv;


    def nrows (self):
        """
        number of rows of the defining set
        """    
        return self._matrix.nrows();

    def ncols (self):
        """
        number of columns of the defining set
        """
        return self._matrix.ncols();          
                 
    def rank (self):
        return self._rank; 

    def dim (self):
        """
        Give the dimension of the cell in the retract.
        dim = 6-rank
        """
        return 6 - self._rank; 

    #=====================================================
    # Compare cells
    #=====================================================

    def __eq__ (self, other):
        if self._matrix == other._matrix:
            return True;
        else:
            return False;

    def __ne__ (self, other):
        if self._matrix != other._matrix:
            return True;
        else:
            return False;


    def __cmp__ (self, other):
        if self == other:
            return 0;
        elif self.is_in_closure_of(other):
            return -1;
        elif other.is_in_closure_of(self):
            return 1;
        else:
            return None;

    def __le__ (self, other):
        return self.is_in_closure_of(other);


    def __ge__ (self, other):
        return other.is_in_closure_of(self);

    #=====================================================
    # actions on the cells
    #=====================================================


    def __mul__ (self, other):
        """
        computes the image of self under the action of a matrix or a matrix group elemenent.
        """
        if is_Matrix(other):
            __matrix = other
        elif is_MatrixGroupElement(other):
            __matrix=other.matrix()
        else: 
            try :
                __matrix = matrix(other)
            except:
                raise TypeError, "One Factor has to be a matrix or a matrix group element";

        #Compute Matrix
        __out = getstandardform((__matrix.inverse())*(self.matrix()));
        #map2std
        if self._map2standard == None:
            __out_map2std = None
        else:
            __out_map2std = self.map2standard()*__matrix;


        # Compute stabilizer
        if self._stab == None:
            __out_stab = None
        else:
            #It is better to compute map to std than later compute the stabilizer
            if __out_map2std == None:
                __out_map2std = getmap2standard(__out)

            _stablst = [__matrix.inverse()*g*__matrix for g in self._stab]
            __out_stab = MatrixGroup(_stablst)
        
        # Closure   
        if self._closure == None:
            __out_closure = None;
        else:
            
            # hs to be remove due to the fact that for neighbour a infinite loop occured
            #__out_closure = [c.apply(__matrix) for c in self._closure];
            __out_closure = None
        __name       = self._name+"_img";
        __latex_name = "{"+self._latex_ +"}_{img}";  
        return Cell (__out, name =  __name, latex_name = __latex_name, cell_rank = self._rank, cell_type = self._cell_type, transv = self._transv, bd_type = self._bd_type,  map2standard = __out_map2std, stabilizer = __out_stab, closure = __out_closure)
        
                
    def __rmul__ (self, other):
        return self*other;
  
        
    def apply(self, other, new_name = None, new_latex = None, check=True):
        """
        computes the image of self under the action of a matrix or 
        a matrix group elemenent.
        
        Remark::
        
        Do the same as __mul__ but has the propertie that one can give 
        the object a new name via new_name = str and new_latex =str.
        """
        if check:
            if is_Matrix(other):
                __matrix = other;
            elif is_MatrixGroupElement(other):
                __matrix=other.matrix();
            else: 
                try :
                    __matrix = matrix(other);
                except:
                    raise TypeError, "One Factor has to be a matrix or a matrix group element";

            if not __matrix.is_invertible():
                raise ZeroDivisionError, "Matrix not invertible";
        __out = self * __matrix;
        if new_name!= None:
            __out.rename (new_name);
        if new_latex!= None:
            __out.rename_latex(new_latex)
        return __out;


    #=====================================================
    # Boundary and Closure
    #=====================================================

    def deleteclosure (self):
        self._closure = None;

    def closure (self, basedir = None, dir = 'output/cells/closures'):
        """
        RETURN:
        List of cells which ly in the boundary of the cell grouped by.
        optional parametr where to search path =
        """
        if self._closure != None:
            return self._closure
        else:
            return getclosure (self,basedir = basedir, dir = dir)
        

    def boundary (self, basedir = BASEPATH, dir =  'output/cells/closures'):
        """
        RETURN:
        The boundary of cell in the cell decoposition, as a list of cells. 
        """
        if self._closure != None:
            return [cell for cell in self._closure if cell != self];
        else:
            _cls = self.closure(basedir = basedir, dir = dir);
            self.set_closure(_cls);
            return [cell for cell in _cls if cell != self];

    def skeleton (self, int, basedir = None, dir = 'output/cells/closures'):
        """
        OUTPUT:
        The n-skeleton of cell in the cell decoposition, as a list of cells.       
        """
        if self._closure != None:
            _tmp = [cell for cell in self._closure if cell.dim() <= int]
        else:
            _cls = self.closure(basedir = basedir, dir = dir)
            self.set_closure(_cls)
            _tmp = [cell for cell in _cls if cell.dim() <= int]
        return _tmp    
        
        




    def is_in_closure_of (self, other):
        """
        Returns True if self is in the closure of other
        """ 

        return all([column in self.matrix().columns() for column in other.matrix().columns()]);

    def is_in_boundary_of (self, other):
        """
        Returns True if self is in the boundary of other
        """ 
        return self.is_in_closure_of (other) and (other != self);
    
    def is_in_direct_boundary_of (self, other):
        """
        Returns if self is boundary of other and difference of dimensions is one
        """ 
        if other.dim() == self.dim()+1:
            print "check direction"
            return self.is_boundary_of (other) ; 
        else:
            return False;
           
    #=====================================================
    # Neighbours
    #=====================================================

    def deleteneighbours (self):
        self._nb = None;

    def neighbours (self, type = None, basedir = None, dir = 'output/cells/neighbours'):
        """
        RETURN:
        List of neighbors of self of type type
        Default is 
        type=None
        which returns neighbors of all types.
        
        
        """
        out = getneighbours(self,type = type, basedir = basedir, dir=dir)
        return out


    #=====================================================
    # Things Related to Groups
    #=====================================================

    def deletestabilizer (self):
        self._stab = None;

    def stabilizer (self, basedir= BASEPATH, dir='output/cells/stabilizers'): 
        """
        cell.stabilizer()
        
        OUTPUT:
        
        - Matrix group of symplectic matrices which stabilizes the cell.
        """
    
        return getstabilizer(self, basedir=basedir, dir=dir)
        
    def stabilizerGSp (self,basedir= BASEPATH, dir='output/cells/stabilizers'): 
        """
        cell.stabilizer()
        
        OUTPUT:
        
        - Matrix group of general symplectic matrices which stabilizes the cell.
        """
    
        return getstabilizerGSp(self, basedir= basedir, dir=dir)        

    def orbit (self, matrix, max_int=500):
        """
        Try to give a list of Elements in the orbit.
        
        It genereated upto max_int=500 (12 is enough) elements by 
        
                matrix^i*self 
                
        and breaks if there is a returning element;
        
        Know that this is a stupid way. 
        """
        
        __out = [];
        __i=0;
        __tmp = self
        while __i < max_int:
            __tmp = matrix*__tmp;
            if __tmp not in __out:
                __out.append(__tmp);
            else:
                print 'found circle';
                return __out;
            
            __i+= 1;
        print 'No circle found with maximum of %S steps increase max_int. Otherwise the list will be incomplete' ;
        return __out;  
          


#===============================================================================
# StandardCell()
#===============================================================================

class StandardCell (Cell):
    """
    Defining the class for defining Standard cells:
    Only used for faster imput of the cells by chosing columns from some collection.
    """
    
    def __init__ (self, vector_index, std_name, std_latex_name, std_cell_rank, std_cell_type, std_transv = None, std_bd_type=False, closure = None, stabilizer = None):
        _std_vectors = matrix(ZZ,4,[1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1,
0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 1,
-1, 0, 1, -1, 0, -1, 1, 0, 1, 0, 0, 0, 1, 0, 0, -1, 1, 0, -1, 0, -1, -1,
0, -1, 1, -1, 1, 0]);
       
        _matrix    = getstandardform(_std_vectors.matrix_from_columns(vector_index));

        

        Cell.__init__(self, _matrix , name = std_name, latex_name = std_latex_name, cell_rank = std_cell_rank, cell_type = std_cell_type, transv = std_transv, bd_type = std_bd_type , map2standard = identity_matrix(4), closure = closure, stabilizer = stabilizer);



#===============================================================================
# PART III: Standard Cells and Cell Lists
#===============================================================================


Desargues = StandardCell ([0,1,2,3,4,5,6,7,8,9],"Desargues0", "\\Des", 6, "Desargues", std_transv = True, std_bd_type = False);
Reye      = StandardCell ([0,1,2,3,4,10,11,12,13,14,15,16], "Reye0", "\\Reye", 6, "Reye", std_transv = True, std_bd_type = False);
DR        = StandardCell ([0,1,2,3,4,5,6,7,9], "DR0", "\\DR", 5, "DR", std_transv = True, std_bd_type = False);
RR        = StandardCell ([0,1,2,3,4,11,12,13,15],"RR0" , "\\RR", 5, "RR", std_transv = True, std_bd_type = False);
Hexagon   = StandardCell ([0,1,2,3,17,18], "Hexagon0", "\\Hex", 4, "Hexagon", std_transv = False, std_bd_type = False);
Square    = StandardCell ([0,1,2,3,4,6,7,9], "Square0", "\\Sq", 4, "Square", std_transv = True, std_bd_type = False);
Triangle  = StandardCell ([0,1,2,3,4,5,7,9], "Triangle0", "\\Tria", 4, "Triangle", std_transv = True, std_bd_type = False);
Vertebra  = StandardCell ([0,1,2,3,17], "Vertebra0", "\\Vertebra", 3, "Vertebra", std_transv = False, std_bd_type = False);
Crystal   = StandardCell ([0,1,2,3,5,7], "Crystal0", "\\Cryst", 3, "Crystal", std_transv = False, std_bd_type = False);
Pyramid   = StandardCell ([0,1,2,3,4,7,9], "Pyramid0", "\\Pyr", 3, "Pyramid", std_transv = True, std_bd_type = False);
RedSquare = StandardCell ([0,1,2,3], "RedSquare0", "\\RedSq", 2, "RedSquare", std_transv = False, std_bd_type = False);

StandardElementList = [Desargues, Reye, DR, RR, Hexagon, Square, Triangle, Vertebra, Crystal, Pyramid, RedSquare];
CellList            = RedSquare.closure()
ReyeList           = [c for c in CellList if c.cell_type()=='Reye']
DesarguesList      = [c for c in CellList if c.cell_type()=='Desargues']
RRList             = [c for c in CellList if c.cell_type()=='RR']
DRList             = [c for c in CellList if c.cell_type()=='DR']
TriangleList       = [c for c in CellList if c.cell_type()=='Triangle']
SquareList         = [c for c in CellList if c.cell_type()=='Square']
HexagonList        = [c for c in CellList if c.cell_type()=='Hexagon']
VertebraList       = [c for c in CellList if c.cell_type()=='Vertebra']
CrystalList        = [c for c in CellList if c.cell_type()=='Crystal']
PyramidList        = [c for c in CellList if c.cell_type()=='Pyramid']

RedSquare0  = CellList[0]
Vertebra0   = CellList[473]
Vertebra1   = CellList[474]
Crystal0    = CellList[477]
Pyramid0    = CellList[481]
Triangle0   = CellList[361]
Triangle59  = CellList[420]
Triangle74  = CellList[435]
Square0     = CellList[297]
Square12    = CellList[309]
Hexagon0    = CellList[293]
Hexagon1    = CellList[294]
DR0         = CellList[77]
DR55        = CellList[132]
DR110       = CellList[187]
DR136       = CellList[213]
RR0         = CellList[253]
RR15        = CellList[268]
RR19        = CellList[272]
RR37        = CellList[290]
Reye0       = CellList[49]
Reye1       = CellList[50]
Reye3       = CellList[52]
Reye5       = CellList[54]
Reye11      = CellList[60]
Reye16      = CellList[65]
Desargues0  = CellList[1]
Desargues30 = CellList[31]
Desargues39 = CellList[40]

NecessaryCells = [RedSquare0,Vertebra0,Crystal0,Pyramid0,Triangle0,Triangle74,Square0,Square12,Hexagon0,DR0,DR110,RR19,RR37,Desargues0,Desargues30,Reye11,Reye3]
