
"""
cellsio - Module to output cells 

AUTHOR:

- JENS PUTZKA (2010-05-21): initial version

DATE:
- 2010-10-12

VERSION:
- 1.5 - 2010-10-12
- 1.0 - 2010-05-21

"""
#*****************************************************************************
#       Copyright (C) 2009 - JENS PUTZKA <putzka@mpim-bonn.mpg.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

#===============================================================================
# Imports
#===============================================================================

from sage.misc.latex import latex
from sage.graphs.all import DiGraph
from sage.graphs.all import Graph
from io import saveobj2file, loadobj
from io import savelist2file
from config import BASEPATH
from cells import is_Cell, StandardElementList


#===============================================================================
# PART I: Functions and Constants
#===============================================================================

#===============================================================================
# celloutput ()
#===============================================================================

def celloutput (cell, style):
    """
    General Output construction:
    INPUT 
    -- cell : can be:
            -- single cell
            -- list of cells in particuar a cell complex
    -- style:
        -- 'screen'   : Returns Matrix as text
        -- 'latex'    : Returns Matrix as latex commmand
        -- 'skeleton' : Returns a graph of the 1-skeleton
        -- 'plot2d'   : Returns the plot of the 1-skeleton
        -- 'plot3d'   : Returns the 3d plot of the 1-skeleton
        -- 'graph'    : Returns the directed graph of the closure of cell, 
                      where the vertices are the cells and there is an 
                      edge if and only if a cell lie in the boundary of 
                      and has codimension 1.
    """
    # Test if cell or cell list
    if not is_Cell(cell):
        
        if isinstance(cell,list):
           if not all(is_Cell(c) for c in cell) :
               raise TypeError, "Input is not a cell or a list of cells.";
           else:
               raise TypeError, "Input is not a cell or a list of cells.";

    # Different cases of display
    if style == 'screen':
        return cell.matrix();
    elif style == 'latex':
        return cell.matrix_latex()
    
    elif style == 'plot2d':
        _tmp       = cell.skeleton(1);
        _tmp_edge = [a for a in _tmp if a.dim()==1];
        _tmp_vert = [b for b in _tmp if b.dim()== 0];
        _tmp_g = Graph({}, loops=False, multiedges=False);
        for e in _tmp_edge:
            
            _tmp = tuple([str(v) for v in _tmp_vert if v.is_in_closure_of(e)])
            _tmp_g.add_edge (_tmp, label=e.cell_type() )        
        return _tmp_g.plot()
    elif style == 'plot3d':
        _tmp       = cell.skeleton(1);
        _tmp_edge = [a for a in _tmp if a.dim()==1];
        _tmp_vert = [b for b in _tmp if b.dim()== 0];
        _tmp_g = Graph({}, loops=False, multiedges=False);
        for e in _tmp_edge:
            
            _tmp = tuple([str(v) for v in _tmp_vert if v.is_in_closure_of(e)])
            _tmp_g.add_edge (_tmp, label=e.cell_type() )        
        return _tmp_g.plot3d()        
        
    elif style == 'skeleton':
        _tmp       = cell.skeleton(1);
        _tmp_edge = [a for a in _tmp if a.dim()==1];
        _tmp_vert = [b for b in _tmp if b.dim()== 0];
        _tmp_g = Graph({}, loops=False, multiedges=False);
        for e in _tmp_edge:
            
            _tmp = tuple([str(v) for v in _tmp_vert if v.is_in_closure_of(e)])
            _tmp_g.add_edge (_tmp, label=e.cell_type() )
            
        
        return _tmp_g;
    elif style == 'graph':
        print "Start\n"
        print "   * Getting Closure"
        _tmp       = cell.closure();
        
        _tmp_vert = [b for b in _tmp];
        _tmp_edge = [[a,b] for a in _tmp for b in _tmp if a.is_in_closure_of(b) and (a.dim()-b.dim()== 1)] ;
        
        _tmp_graph = DiGraph({}, loops=False, multiedges=False);
        for e,f in _tmp_edge:
            
            _tmp_graph.add_edge ([str(e),str(f)],label=e.cell_type() )
        
        return _tmp_graph
                    
    else:
        print "The style %s is unknown. None Returned" %style; 
        return None;

#===============================================================================
# savecell2file ()
#===============================================================================



def savecell2file (cell,style, filetype, filename='out',path = None):
    """
    Same as celloutput see there for more options but print it to a file
    
    INPUT
    --cell     : as cellprint
    --style    : 
        -- 'screen'   : Returns Matrix as text
        -- 'latex'    : Returns Matrix as latex commmand
        -- 'plot2d'   : Filetype has to be one of the following eps, pdf, png
                        ps, svg, sobj(=Sage object)
        -- 'plot3d'   :
        -- 'skeleton' :
    
    --filetype : tex/jpg/png/pdf/ps/txt 
    --filename : default out
    --path     : 
    """
    if path == 'mpi':
        _path='/home/putzka/Dissertation/Latex/sage_files/saved/';
    elif path == 'home':
        _path='/Users/jens/Dropbox/Dissertation/sage_files/saved/';
    elif path == 'laptop':
        _path='/home/jens/Dropbox/Dissertation/sage_files/saved/';
    else:
        _path= path;
    
    __out = celloutput (cell, style)


    save2file(__out,filename, basedir = _path, dir= 'output' ,extension=filetype)
        
        
 #===============================================================================
# savecelllist2file ()
#===============================================================================    
    
def savecelllist2file (cell_list,style, filetype='txt', filename='out',path = BASEPATH):
    """
    Same as cellprint see there for more options but print it to a file
    
    INPUT
    --cell     : as cellprint
    --style    : 
        -- 'screen'   : Returns Matrix as text
        -- 'latex'    : Returns Matrix as latex commmand
        -- 'latex_table_cell' : 
        -- 'latex_table_stab' : 
        -- 'latex_table_nb' : 
        -- 'plot2d'   : 
        -- 'plot3d'   :
        -- 'skeleton' :
    
    --filetype : tex/jpg/png/pdf/ps/txt default=txt
    --filename : default out.txt
    --path     : 
    """
    if path == 'mpi':
        _path='/home/putzka/Dropbox/Dissertation/sagefiles/saved/';
    elif path == 'home':
        _path='/Users/jens/Dropbox/Dissertation/sagefiles/saved/';
    elif path == 'laptop':
        _path='/home/jens/Dropbox/Dissertation/sagefiles/saved/';
    else:
        _path= path;

    _out_file = open(_path+filename+'.'+filetype, 'w');

    if style == 'screen':
        pass;
    elif style == 'latex':
        for cell in cell_list:
            _out_file.write(cell.latex());
    elif style == 'latex_matrices':
        for cell in cell_list:
            _out_file.write(cell.matrix_latex());
            _out_file.write('\n');
    elif style == 'latex_table_cell':
        _out_file.write('\\begin{longtable}{rll}\n{\\em Name}&{\\em Cell Matrix}&{\\em Map to Std.Element}\\\\\n\\hline\\\\[0.5em]\n\endhead\n'); 
        for cell in cell_list: 
            _out_file.write('$');
            _out_file.write(cell._latex_);
            _out_file.write('$&$');
            _out_file.write(cell.matrix_latex());
            _out_file.write('$&$');
            _out_file.write(latex(cell.map2standard()));
            _out_file.write('$\\\\\n');
        _out_file.write('\\end{longtable}\n'); 
    elif style == 'latex_table_stab':
        _out_file.write('\\begin{longtable}{rl}\n{\\em Name}&{\\em Stabilizer Elements}\\\\\n\\hline\\\\[0.5em]\n\endhead\n'); 
        for cell in cell_list: 
            _out_file.write('$');
            _out_file.write(cell._latex_);
            _out_file.write('$&$');
            _out_file.write(latex(cell.stabilizer()));
            _out_file.write('$\\\\\n');
        _out_file.write('\\end{longtable}\n'); 
    elif style == 'latex_table_nb':
        _out_file.write('\\begin{longtable}{rl}\n{\\em Name}&{\\em Type Neighbours}&{\\em Neighbours}\\\\\n\\hline\\\\[0.5em]\n\endhead\n'); 
        for cell in cell_list: 
            _out_file.write('$');
            _out_file.write(cell._latex_);
            _out_file.write('$');
            _out_file.write('&');
            flg = 0
            for ctp  in StandardElementList:
                if ctp.dim()<= cell.dim():
                    continue
                _out_file.write('$');
                _out_file.write(latex(ctp.cell_type()));
                _out_file.write('$');
                _out_file.write('&');
                for c in cell.neighbours():
                    if c.cell_type() ==ctp.cell_type():
                        flg =flg+1
                        _out_file.write('$');
                        _out_file.write(c.matrix_latex());
                        _out_file.write('$');
                        _out_file.write(',');
                _out_file.write('\\\\\n');
                if flg > 0:
                    _out_file.write('&');
            _out_file.write('\\\\\n');
        _out_file.write('\\end{longtable}\n');     
    elif style == 'plot2d':
        pass;
    elif style == 'plot3d':
        pass;
    elif style == 'skeleton':
        pass;
    else:
        print "The style %s is unknown. None Returned" %style; 
        return None;
    _out_file.close();
    print 'Done';       

#===============================================================================
# PART II: Classes
#===============================================================================
