
"""
HighestWeightModule - Module to decribe weight modules and actions on them using the GAP wrapper

AUTHOR:

- JENS PUTZKA (2009-09-10): Initial version
- JENS PUTZKA (2011-10-10): Final version

DATE:
- 2010-10-06

VERSION:
- 1.2 - 2011-10-10
- 1.1 - 2011-06-07
- 1.0 - 2011-02-03
- 0.6 - 2011-01-19


EXAMPLES::

USAGE::

REFERENCES::

"""
#*****************************************************************************
#       Copyright (C) 2012 - JENS PUTZKA <putzka@mpim-bonn.mpg.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

print "\n=========================================================";
print '|  Importing HighestWeightModule Package  (c) 2011      |';
print '|                                                       |'
print '|    by Jens Putzka MPI for Mathematics Bonn            |'
print "=========================================================";

#===============================================================================
# Imports
#===============================================================================
from symplectic import symplectic2symplectic

from sage.interfaces.gap import gap
from sage.rings.rational_field import QQ
from sage.rings.integer_ring import ZZ
from sage.misc.prandom import randint 
from sage.categories.monoids import prod
from sage.matrix.all import matrix, is_Matrix, identity_matrix
from matrix_operations import is_diagonal_matrix, expmat 
from  sage.groups.all import MatrixGroup, is_MatrixGroupElement, is_MatrixGroup
from sage.modules.all import VectorSpace


gap.LoadPackage('"quagroup"')

#===============================================================================
# SimpleLieAlgebra
#===============================================================================

def SimpleLieAlgebra(type,rank, base_ring=QQ):
    """
    INPUT::
        type = 'A','B','C',... Cartan type
        rank = int valid rank for that type
        base_ring = QQ (default), ring: ZZ,QQ,RR,...
    OUTPUT::
        Simple Lie Algebra object coming from GAP library
    
    
    EXAMPLE::
        sage: L = SimpleLieAlgebra('C',2);L
              Algebra( Integers, [ v.1, v.2, v.3, v.4, 
                         v.5, v.6, v.7, v.8, v.9, v.10 ])
        sage: L.Basis()[1]
                v.1
        Take care indices of list coming from GAP start 
        with 1 and not with 0 as usual SAGE/PYTHON lists. 
        
        sage: L.Dimension()
                10 
        sage: L.DimensionOfHighestWeightModule([10,3])
                2090
        sage: R = L.RootSystem()
                <root system of rank 2>
        sage: R.PositiveRoots()
                [ [ 2, -1 ], [ -2, 2 ], [ 0, 1 ], [ 2, 0 ] ]

    """
    gap.LoadPackage('"quagroup"')
    _type ='"%s"'%type
    return gap.SimpleLieAlgebra(_type, rank, base_ring)

#===============================================================================
# LieType
#===============================================================================

def LieType(L):
    """
    INPUT::
        L : SimpleLieAlgebra
    RETURN::
        list = [CartanType,rank]
    EXAMPLES::
        sage: L = SimpleLieAlgebra ("C", 2);
        sage: LieType (L)
            ['C', 2]
    """
    if not gap.IsSimpleAlgebra(L):
        raise TypeError, "L has to be a simple algebra";
    _out = str(gap.SemiSimpleType(L))
    #print type(_out)
    return [_out[0],int(_out[1:])]

#===============================================================================
# HighestWeightModule
#===============================================================================

def HighestWeightModule(L, weight):
    """
    INPUT::
        L      = SimpleLieAlgebra object
        weight = list of non negative integers of length rank of L
    OUTPUT::
        Highest weight module over L for the highest weight weight
        coming from a GAP object
    EXAMPLE::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: M = HighestWeightModule(L,[5,2]); M
            <324-dimensional left-module over Algebra( Rationals, 
            [ v.1, v.2, v.3, v.4, v.5, v.6, v.7, v.8, v.9, v.10 ] )>
    """
    
    #if not gap.LoadPackage('"quagroup"'):
    gap.LoadPackage('"quagroup"')    
    
    
    return L.HighestWeightModule(weight)
    #return gap.HighestWeightModule(L,weight)

#===============================================================================
# latticegeneratorsinUEA
#===============================================================================

def latticegeneratorsinUEA (L):
    """
    INPUT::
        L      = Simple Lie Algebra
    OUTPUT::
        Kostant Lattice generators for the universal enveloping algebra 
        as SAGE list 
    EXAMPLE::
        sage: SimpleLieAlgebra('C',2) 
        sage: latticegeneratorsinUEA (L)
            [y1, y2, y3, y4, x1, x2, x3, x4, ( h9/1 ), ( h10/1 )]
    """    
    
    gap.LoadPackage('"quagroup"')
    return [u for u in L.LatticeGeneratorsInUEA()]

#===============================================================================
# actionmatrixofUEAelement
#===============================================================================
    
def actionmatrixofUEAelement(UEAel,HWM, base_ring=QQ, sparse=True):
    """
    INPUT::
        UAEel     = Element in the universal enveloping algebra
        HWM       = Highest Weight Module
        base_ring = QQ (default)
        sparse    = True (default)
    OUTPUT::
        matrix of action of UAEel on HWM in the basis HWM.Basis()
        
    EXAMPLE::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: M = HighestWeightModule( L,[0,1]); M
        <5-dimensional left-module over Algebra( Rationals, 
        [ v.1, v.2, v.3, v.4, v.5, v.6, v.7, v.8, v.9, v.10 ] )>
        sage: Ugen = latticegeneratorsinUEA(L)
        sage: actionmatrixofUEAelement( Ugen[1], M)
            [0 0 0 0 0]
            [1 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 1 0]
    """

    gap.LoadPackage('"quagroup"')    
    
    entries = dict();
    _B = [gap.ExtRepOfObj(b) for b in HWM.Basis()]
    _dim = len(_B)
    _Null = (0*_B[0]).ExtRepOfObj()
    j=0
    for b in _B:
        #z = UEAel^b
        z = gap.POW(UEAel,b)
        z_ = z.ExtRepOfObj()
        
        if z_!=_Null:
            i = 1
            _len = len(z_)
            while i < _len: 
                entries[(int(z_[i,1])-1,j)] = int(z_[i+1])
                i+=2
        j+=1
    return matrix(base_ring, _dim, entries, sparse=sparse)


#===============================================================================
# genericalgebragenerators
#===============================================================================

def genericalgebragenerators (L, base_ring=QQ):
    """
    INPUT::
        L = SimpleLieAlgebra
    OUTPUT::
        Nilpotent generators of L coming from the root system 
        of L as matrices
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: genericalgebragenerators(L)
        [
        [0 0 0 0]  [ 0  0  0  0]  [ 0  0  0  0]  [0 0 0 0]  [0 1 0 0]
        [1 0 0 0]  [ 0  0  0  0]  [ 0  0  0  0]  [0 0 0 0]  [0 0 0 0]
        [0 0 0 0]  [ 0 -1  0  0]  [ 1  0  0  0]  [0 0 0 0]  [0 0 0 1]
        [0 0 1 0], [ 0  0  0  0], [ 0 -1  0  0], [1 0 0 0], [0 0 0 0],

        [ 0  0  0  0]  [ 0  0  1  0]  [0 0 0 1]
        [ 0  0 -1  0]  [ 0  0  0 -1]  [0 0 0 0]
        [ 0  0  0  0]  [ 0  0  0  0]  [0 0 0 0]
        [ 0  0  0  0], [ 0  0  0  0], [0 0 0 0]
        ]
    """
    
    gap.LoadPackage('"quagroup"')  
    
    _R      = L.RootSystem()
    _N      = len(_R.NegativeRoots())
    _Ugen   = latticegeneratorsinUEA(L)
    _rank   = LieType(L)[1]
    _wgt    = [0 for i in range(_rank)]
    _wgt[0] = 1
    _V      = HighestWeightModule(L,_wgt)
    
    
    _neg_mat = [actionmatrixofUEAelement(_Ugen[i],_V, base_ring=base_ring)  for i in range(_N)]
    _pos_mat = [actionmatrixofUEAelement(_Ugen[i+_N],_V, base_ring=base_ring)  for i in range(_N)]
    return _neg_mat+_pos_mat

#===============================================================================
# negativegenericalgebragenerators
#===============================================================================

def negativegenericalgebragenerators (L, base_ring=QQ):
    """
    INPUT::
        L = SimpleLieAlgebra
    OUTPUT::
        Nilpotent generators of L coming from the negative roots 
        in the root system of L as matrices
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: negativegenericalgebragenerators(L)
                [
                [0 0 0 0]  [ 0  0  0  0]  [ 0  0  0  0]  [0 0 0 0]
                [1 0 0 0]  [ 0  0  0  0]  [ 0  0  0  0]  [0 0 0 0]
                [0 0 0 0]  [ 0 -1  0  0]  [ 1  0  0  0]  [0 0 0 0]
                [0 0 1 0], [ 0  0  0  0], [ 0 -1  0  0], [1 0 0 0]
                ]

    """
    
    gap.LoadPackage('"quagroup"')  
    
    _R      = L.RootSystem()
    _N      = len(_R.PositiveRoots())
    _Ugen   = latticegeneratorsinUEA(L)
    _rank   = len(_Ugen)-2*_N
    _wgt    = [0 for i in range(_rank)]
    _wgt[0] = 1
    _V      = HighestWeightModule(L,_wgt)
    
    
    _neg_mat = [actionmatrixofUEAelement(_Ugen[i],_V, base_ring=base_ring)  for i in range(_N)]
    #_pos_mat = [actionmatrixofUEAelement(_Ugen[i+_N],_V, base_ring=base_ring)  for i in range(_N)]
    return _neg_mat

#===============================================================================
# positivegenericalgebragenerators
#=============================================================================== 

def positivegenericalgebragenerators (L, base_ring=QQ):
    """
    INPUT::
        L = SimpleLieAlgebra
    OUTPUT::
        Nilpotent generators of L coming from the positive roots 
        in the root system of L as matrices
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: positivegenericalgebragenerators(L)
                [
                [0 1 0 0]  [ 0  0  0  0]  [ 0  0  1  0]  [0 0 0 1]
                [0 0 0 0]  [ 0  0 -1  0]  [ 0  0  0 -1]  [0 0 0 0]
                [0 0 0 1]  [ 0  0  0  0]  [ 0  0  0  0]  [0 0 0 0]
                [0 0 0 0], [ 0  0  0  0], [ 0  0  0  0], [0 0 0 0]
                ]
    """
    
    gap.LoadPackage('"quagroup"')  
    
    _R      = L.RootSystem()
    _N      = len(_R.PositiveRoots())
    _Ugen   = latticegeneratorsinUEA(L)
    _rank   = len(_Ugen)-2*_N
    _wgt    = [0 for i in range(_rank)]
    _wgt[0] = 1
    _V      = HighestWeightModule(L,_wgt)
    
    
    #_neg_mat = [actionmatrixofUEAelement(_Ugen[i],_V, base_ring=base_ring)  for i in range(_N)]
    _pos_mat = [actionmatrixofUEAelement(_Ugen[i+_N],_V, base_ring=base_ring)  for i in range(_N)]
    return _pos_mat

#===============================================================================
# genericgroupgenerators
#=============================================================================== 

def genericgroupgenerators (L, base_ring=QQ):
    """
    INPUT::
        L         = SimpleLieAlgebra
        base_ring = QQ (default)
    OUTPUT::
    OUTPUT::
        Unipotent generators which are exp of genericalgebragenerators
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: genericgroupgenerators(L)
                [
                [1 0 0 0]  [ 1  0  0  0]  [ 1  0  0  0]  [1 0 0 0]  [1 1 0 0]
                [1 1 0 0]  [ 0  1  0  0]  [ 0  1  0  0]  [0 1 0 0]  [0 1 0 0]
                [0 0 1 0]  [ 0 -1  1  0]  [ 1  0  1  0]  [0 0 1 0]  [0 0 1 1]
                [0 0 1 1], [ 0  0  0  1], [ 0 -1  0  1], [1 0 0 1], [0 0 0 1],

                [ 1  0  0  0]  [ 1  0  1  0]  [1 0 0 1]
                [ 0  1 -1  0]  [ 0  1  0 -1]  [0 1 0 0]
                [ 0  0  1  0]  [ 0  0  1  0]  [0 0 1 0]
                [ 0  0  0  1], [ 0  0  0  1], [0 0 0 1]
                ]
    """
    
    return [expmat(g) for g in genericalgebragenerators(L, base_ring=base_ring)]

#===============================================================================
# positivegenericgroupgenerators
#=============================================================================== 

def positivegenericgroupgenerators (L, base_ring=QQ):
    """
    INPUT::
        L         = SimpleLieAlgebra
        base_ring = QQ (default)
    OUTPUT::
    OUTPUT::
        Unipotent generators which are exp of genericalgebragenerators
        coming from positive roots.
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: positivegenericgroupgenerators (L)
           [
           [1 1 0 0]  [ 1  0  0  0]  [ 1  0  1  0]  [1 0 0 1]
           [0 1 0 0]  [ 0  1 -1  0]  [ 0  1  0 -1]  [0 1 0 0]
           [0 0 1 1]  [ 0  0  1  0]  [ 0  0  1  0]  [0 0 1 0]
           [0 0 0 1], [ 0  0  0  1], [ 0  0  0  1], [0 0 0 1]
           ] 
    """
    return [expmat(g) for g in positivegenericalgebragenerators(L,base_ring=base_ring)]

#===============================================================================
# negativegenericgroupgenerators
#=============================================================================== 

def negativegenericgroupgenerators (L, base_ring=QQ):
    """
    INPUT::
        L         = SimpleLieAlgebra
        base_ring = QQ (default)
    OUTPUT::
    OUTPUT::
        Unipotent generators which are exp of genericalgebragenerators
        coming from negative roots.
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: positivegenericgroupgenerators (L)
           [
           [1 0 0 0]  [ 1  0  0  0]  [ 1  0  0  0]  [1 0 0 0]
           [1 1 0 0]  [ 0  1  0  0]  [ 0  1  0  0]  [0 1 0 0]
           [0 0 1 0]  [ 0 -1  1  0]  [ 1  0  1  0]  [0 0 1 0]
           [0 0 1 1], [ 0  0  0  1], [ 0 -1  0  1], [1 0 0 1]
           ]
    """
    
    return [expmat(g) for g in negativegenericalgebragenerators(L,base_ring=base_ring)]

#===============================================================================
# genericgroupgeneratorsfunctions
#=============================================================================== 

def genericgroupgeneratorsfunctions (L, base_ring=QQ):
    """
    INPUT::
        L         = SimpleLieAlgebra
        base_ring = QQ (default)
    OUTPUT::
        Unipotent generators as functions which are exp(t*z) 
        for z in genericalgebragenerators
    EXAMPLES::
        sage: L = SimpleLieAlgebra('C',2) 
        sage: f = genericgroupgeneratorsfunctions (L)[0]
        sage f(8)
            [1 0 0 0]
            [8 1 0 0]
            [0 0 1 0]
            [0 0 8 1]
    """
    
    return map(lambda t: (lambda x: expmat(t*x)), genericalgebragenerators(L, base_ring=base_ring))

#===============================================================================
# genericgroupgeneratorsfunctionsonHWM
#=============================================================================== 

def genericgroupgeneratorsfunctionsonHWM (HWM, base_ring=QQ,sparse=True, UEA_generators= []):
    """
    INPUT::
        
        base_ring = QQ (default)
    OUTPUT::
        Unipotent generators as functions which are exp(t*z) 
        for z in genericalgebragenerators
    """
    
    gap.LoadPackage('"quagroup"')  
    L    = HWM.LeftActingAlgebra()

    if UEA_generators == []:
        UEA_generators = latticegeneratorsinUEA(L)

    _R      = L.RootSystem()
    _N      = len(_R.NegativeRoots())
    #_Ugen   = latticegeneratorsinUEA(L)
   
    
    
    
    gens =  [actionmatrixofUEAelement(UEA_generators[i],HWM, base_ring=base_ring)  for i in range(2*_N)]
    return map(lambda t: (lambda x: expmat(t*x)), gens)


#===============================================================================
# positivegenericgroupgeneratorsfunctions
#=============================================================================== 

def positivegenericgroupgeneratorsfunctions (L, base_ring=QQ):
    """
    INPUT::
        L         = SimpleLieAlgebra
        base_ring = QQ (default)
    OUTPUT::
        Unipotent generators as functions which are exp(t*z) 
        for z in positivegenericalgebragenerators
    """
    
    return map(lambda t: (lambda x: expmat(t*x)), positivegenericalgebragenerators(L, base_ring=base_ring))

#===============================================================================
# negativegenericgroupgeneratorsfunctions
#=============================================================================== 
    
def negativegenericgroupgeneratorsfunctions (L, base_ring=QQ):
    """
    INPUT::
        L         = SimpleLieAlgebra
        base_ring = QQ (default)
    OUTPUT::
        Unipotent generators as functions which are exp(t*z) 
        for z in negativegenericalgebragenerators
    """
    return map(lambda t: (lambda x: expmat(t*x)), negativegenericalgebragenerators(L, base_ring=base_ring))
    
#===============================================================================
# div_min
#===============================================================================    
def div_min(div,n):
    """
    div_min::
        
        INPUT::
        a,b with abs(a)<= abs(b)
    OUTPUT::
        [s,t]: t=min(abs(s*a+b))
    EXAMPLE::
        sage: div_min(8,9)
            [-1, 1]
        sage: div_min(7,90)
            [-13, -1]
    """
    try:
        div=int(div);
        n=int(n)
    except:
        try:
            div=float(div);
            n=float(n);
        except:
            raise ValueError, "Arguments of div_min has to be integers or floats"
    if div== 0:
        raise ZeroDivisionError, "Division by zero"
    if n==0:
        return [0,0]
    if abs(n)<abs(div):
        raise ValueError, "div>n";
    _out= divmod(n,div)
    _out = [-1*_out[0],_out[1]]
    
    
    if abs(_out[1] )> abs(div/2):
        _out = [_out[0]-1,_out[1]-div];
        
    return _out

#===============================================================================
# sign
#===============================================================================

def sign (n):
    """
    INPUT::
        n : integer
    OUTPUT::
        n/abs(n)
    """
    return n/abs(n)
    
#===============================================================================
# is_zeroed
#===============================================================================
def is_zeroed(mat, step=0):
    """
    Test whether all entries below the step-th diagonal entry 
    are zero
    INPUT::
        mat
        step = 0 (i.e. column and row)
    """
    
    _col = mat.column(step)
    _n   = len (_col)
     
    return all([_col[i]==0 for i in range(step+1,_n)])
    
    
#===============================================================================
# genericpermutations
#=============================================================================== 

def genericpermutations(L,negative_list=[],positive_list=[], base_ring=QQ):
    """
    INPUT::
        negative_list (optional) = list of genericgroupgeneratorsfunctions coming from neg. roots
        positive_list (optional) = list of genericgroupgeneratorsfunctions coming from pos. roots
    OUTPUT::
        List of [i,mat,perm] with mat=y_i(1)*x_i(-1)*y_i(1) and perm is the permutation defined by mat as list of tuples.
    EXAMPLES::
        sage: L=SimpleLieAlgebra('C',2)
        sage: P=genericpermutations(L)
        sage: P[0]
            [0, [ 0 -1  0  0]
                [ 1  0  0  0]
                [ 0  0  0 -1]
                [ 0  0  1  0], [(0, 1), (2, 3)]]
    """
    if negative_list==[]:
        negative_list=negativegenericgroupgeneratorsfunctions(L, base_ring=base_ring)
    if positive_list==[]:
        positive_list=positivegenericgroupgeneratorsfunctions(L, base_ring=base_ring)
    
    
    _len = len (negative_list)
    _dim = negative_list[0](1).ncols(); 
    _out = [[i,negative_list[i](1)*positive_list[i](-1)*negative_list[i](1)] for i in range(_len)]
    _out = [[o_i, o_m , [(i,o_m.nonzero_positions_in_column(i)[0]) for i in range(_dim)]] for o_i,o_m in _out ]
    _out = [[o_i,o_m, [(i,j) for i,j  in o_p if i<j]] for o_i, o_m, o_p in _out]
    return _out


#===============================================================================
# append_permutation
#===============================================================================

def append_permutation(L,collection, permutation_index, negative_list=[],positive_list=[],base_ring=QQ):
    """
    INPUT::
        L=
        collection=
        permutation_index =
        permutation_list (optional) =
        negative_list=
        positive_list=
        base_ring =
    OUTPUT::
        None
    """
    
    if negative_list == []:
        negative_list = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list == []:
        positive_list = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
            
    _i = permutation_index
    _n = len(negative_list)
    
    collection.append([[_i,1], negative_list[_i](1)])
    collection.append([[_i+_n,-1], positive_list[_i](-1)])
    collection.append([[_i,1], negative_list[_i](1)])

#===============================================================================
# genericrowoperationdict
#===============================================================================    
    
def genericrowoperationdict (L,negative_list=[], base_ring=QQ):
    """
    (i,j) adds a multiple of the j-th row to the i-th row
    we need those which add multiples of the 1st, i.e. 0th, row to all others in the right order.
    INPUT::
        L                        = SimpleLieAlgebra 
        negative_list (optional) = negativegenericgroupgeneratorsfunctions list of matrices
    OUTPUT::
        dict of (i,j): idx where idx is the index of the element performing the operation with rows i and j.
    EXAMPLE::
        sage: L=SimpleLieAlgebra('C',2)
        sage: genericrowoperationdict(L)
             [{}, {(1, 0): 0}, {(2, 0): 2, (2, 1): 1}, {(3, 0): 3}]
    """
    
    if negative_list==[]:
        negative_list=negativegenericgroupgeneratorsfunctions(L, base_ring=base_ring)
        
    _len = len(negative_list)
    _out = [[(i,j) for i,j in g(1).nonzero_positions() if i!=j] for g in negative_list]
    
    _out = [[idx,_out[idx]] for idx in range(_len)]
    _out2= [[] for k in range(_len)]
    for idx, p in _out:
        _out2[p[0][0]].append([p[0],[idx,negative_list[idx](1)[p[0]] ]])
    

    _out2 = [dict(entr) for entr in _out2]
    return _out2


    
#===============================================================================
# negative_identity
#===============================================================================    
    
def negative_identity(L, step=0, sparse=True, permutation_list=[], negative_list=[], positive_list=[], base_ring=QQ):
    """
    INPUT::
        L
        step
        ncols
        sparse
        permutation_list
    OUTPUT::
        Returns matrix mit +-1 on the diagonal which is equal 
        to -id on the block corresponding to the actual step.
    EXAMPLE::
        sage: L=SimpleLieAlgebra('C',2)
        sage: negative_identity(L)
            [-1  0  0  0]
            [ 0 -1  0  0]
            [ 0  0 -1  0]
            [ 0  0  0 -1]
        sage: negative_identity(L, step= 1)
            [ 1  0  0  0]
            [ 0 -1  0  0]
            [ 0  0 -1  0]
            [ 0  0  0  1]
        sage: L=SimpleLieAlgebra('B',3)
        sage: negative_identity(L)
            [-1  0  0  0  0  0  0]
            [ 0 -1  0  0  0  0  0]
            [ 0  0  1  0  0  0  0]
            [ 0  0  0  1  0  0  0]
            [ 0  0  0  0  1  0  0]
            [ 0  0  0  0  0 -1  0]
            [ 0  0  0  0  0  0 -1]
    """
    
    if permutation_list == []:
        if negative_list == []:
            negative_list = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
        if positive_list == []:
            positive_list = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
        permutation_list = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    
    _ncols = permutation_list[0][1].ncols()
    
    _tmp = [[[i,j],g1*g2] for i,g1,p1 in permutation_list for j,g2,p2 in permutation_list]# 
    for i in range(len(_tmp)):
        if all([_tmp[i][1][j,j]==1 for j in range(step)]) and _tmp[i][1][step,step]==-1:
            #print "Case 1";
            return _tmp[i][1]
    _tmp =[(idx1+idx2, g1*g2) for idx1,g1 in _tmp for idx2,g2 in _tmp  if is_diagonal_matrix(g1*g2)]    
    for i in range(len(_tmp)):
        if all([_tmp[i][1][j,j]==1 for j in range(step)]) and _tmp[i][1][step,step]==-1:
            #print "Case 2";
            return _tmp[i][1]    

    
#===============================================================================
# append_negative_identity
#===============================================================================    

def append_negative_identity(L,collection, step=0, sparse=True, permutation_list=[],negative_list=[],positive_list=[],base_ring=QQ):
    """
    INPUT::
        L
        collection
        step
        ncols
        sparse
        permutation_list
    OUTPUT::
        None : appends generators for sign change to collection.
    """

    
    if negative_list == []:
        negative_list = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list == []:
        positive_list = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if permutation_list == []:
        permutation_list = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)

    nid = negative_identity (L,step=step, sparse=sparse, base_ring=base_ring, permutation_list = permutation_list, negative_list=negative_list, positive_list=positive_list) 
    
    _tmp = [[[i,j],g1*g2] for i,g1,p1 in permutation_list for j,g2,p2 in permutation_list if is_diagonal_matrix(g1*g2)]
    
    _tmp2 = [[idx,g] for idx,g in _tmp if g==nid]
    if len (_tmp2) >= 1:
        _out = _tmp2[0]
    else:
        _out = [(idx1+idx2, g1*g2) for idx1,g1 in _tmp for idx2,g2 in _tmp if g1*g2==nid]
        
        if len(_out)< 1:
            raise NotImplementedError, "Sign combination not found"
        else:
            _out=_out[0]

    for i in _out[0]:
        
        append_permutation (L, collection, i, negative_list=negative_list, positive_list=positive_list, base_ring=base_ring)
        


#===============================================================================
# rowpermutation
#===============================================================================


def rowpermutation(mat,L, step=0, collection=[], negative_list=[], positive_list=[],permutation_list=[], base_ring=QQ, sparse=True):
    """
    Permute rows with symplectic transformations such that at pos (step=0 (i.e. column), row=0)
    is a non zero, positive element.
    
    INPUT::
        L:
        mat :
        step
        ...
    OUTPUT::
        list : [transformed_matrix, collection_of_generators]
    """
    
    _out=mat
    

    if permutation_list == []:
        if negative_list == []:
            negative_list = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
        if positive_list == []:
            positive_list = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    
        permutation_list = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    # Permutations
    
    _collect   = collection[:]
    _len       = len(permutation_list)
    _perm_dict = dict([[p_j,i] for i,m,p in permutation_list for p_j in p])
    
    
    _col      = mat.column(step)
    _min      = 0
    
    _cntr     = 0
    _cntr_min = 0
    for c_i in _col[step:]:
        if c_i!= 0:
            if _min==0 or abs(c_i)<abs(_min):
                _min      = c_i
                _cntr_min = _cntr
        _cntr+=1
    _p   = (step,step+_cntr_min)
    if _p == (step,step):
        if _out[step,step] < 0:
            _out = negative_identity(L, step=step, sparse=sparse, permutation_list=permutation_list, negative_list=negative_list, positive_list=positive_list, base_ring=base_ring)*_out
            append_negative_identity(L,_collect, step=step, sparse=sparse, permutation_list=permutation_list,negative_list=negative_list,positive_list=positive_list,base_ring=base_ring)
            
        return [_out, _collect]
    _i = _perm_dict[_p]
    _perm_mat = permutation_list[_i][1]
    _out = _perm_mat * _out

    _collect.append([[_i, -1], negative_list[_i](-1)])
    _collect.append([[_i+_len, 1], positive_list[_i](1)])
    _collect.append([[_i, -1], negative_list[_i](-1)])
    
    #print " Give the right sign";

    if _out[step,step] < 0:
        _out = negative_identity(L, step=step, sparse=sparse, permutation_list=permutation_list, negative_list=negative_list, positive_list=positive_list, base_ring=base_ring)*_out
        append_negative_identity(L,_collect, step=step, sparse=sparse, permutation_list=permutation_list,negative_list=negative_list,positive_list=positive_list,base_ring=base_ring)
    
    return [_out, _collect]
    
#===============================================================================
# rowtransformation
#===============================================================================

def rowtransformation (mat,L, step=0,  collection=[], negative_list=[],  positive_list=[], permutation_list=[],row_position_list=[], base_ring=QQ, sparse=True):
    """
    INPUT::
        mat        : a matrix
        L           : SimpleLieAlgebra
        collection : set, to which the transformations are appended
    OUTPUT::
        Returns [mat_new,collection_new], where mat_new is a row transfered image of mat
        and a new collection_new, where the additional transformations are 
        appended to collection.
 
    """
    # Generate posiibly non existing lists
    if negative_list     == []:
        negative_list     = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list     == []:
        positive_list     = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if permutation_list  == []:
        permutation_list  = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    if row_position_list == []:
        row_position_list = genericrowoperationdict(L, negative_list=negative_list)

    _collect = collection[:]
    _out_mat = mat
    _len     = mat.nrows()
    _len_n   = len(negative_list)
    
    
    for i in range(step+1,_len ):
        _a   = _out_mat[step,step]
        _n   = _out_mat[i,step]
        
        

        if _n == 0:
            continue;

        _sgn   = row_position_list[i][(i,step)][1]
        if abs(_sgn)!=1:
            raise ValueError, "Entries in generators differnt from +-1";
        _cntr = _sgn*( div_min(_a,_n)[0])

        
        _i       = row_position_list[i][(i,step)][0]

        _row_mat = negative_list[_i](_cntr)

        _out_mat     = _row_mat*_out_mat

        _collect.append([[_i,-_cntr],negative_list[_i](-_cntr)])
        _out     = [_out_mat,_collect]
        
        
        # Permute Elements
        _out = rowpermutation (_out[0], L,step=step, collection=_out[1], negative_list=negative_list, positive_list=positive_list, permutation_list=permutation_list, base_ring = base_ring, sparse=sparse)
        
        _collect = _out[1]
        _out_mat = _out[0]
        
    return _out


#===============================================================================
# randomgroupelemenetbyLiealgebra ()
#===============================================================================

def randomgroupelemenetbyLiealgebra (L, number_of_generators=20, list_of_generators=[], base_ring=ZZ, max=1, min=-1):
    """
    INPUT::
        L                    = SimpleLieAlgebra
        number_of_generators = 10
        list_of_generators   = genericgroupgeneratorsfunctions
        base_ring            = QQ
        max                  =  1
        min                  = -1
    OUTPUT::
        A matrix in the Lie group associated to L, constructed from generators.
        ...
    EXAMPLE::    
       sage: L=SimpleLieAlgebra('C',2)
       sage: randomgroupelemenetbyLiealgebra (L)
           [ 0  3  2  2]
           [ 1  1  0  1]
           [-2  1  2 -1]
           [-1  1  1  0]
    """
    if list_of_generators==[]:
        list_of_generators=genericgroupgeneratorsfunctions(L, base_ring = base_ring)
    _n=len(list_of_generators)
    if base_ring!=QQ:
        _lst = [list_of_generators[randint(0,_n-1)](base_ring.random_element(min,max+1)) for i in  range(number_of_generators)]
    else:
        print "Caution in QQ random_elements behave different"
        [list_of_generators[randint(0,_n-1)](base_ring.random_element(min,max)) for i in  range(number_of_generators)]
    mat = prod(_lst)
    return mat
    
#===============================================================================
# getdecomposition ()
#===============================================================================    
    
def getdecomposition (mat, L, collection=[], negative_list=[], positive_list=[], permutation_list=[],row_position_list=[], base_ring=QQ, sparse=True):
    """
    INPUT::
        mat           = matrix
        L             = SimpleLieAlgebra
        collection    = list to which the elements ar added

    OUTPUT::
        A list, which describes the decompsition of mat in generators.


    """
    if not is_Matrix(mat):
        if not is_MatrixGroupElement(mat):
            raise TypeError, "Object to decompose has to be a matrix or a matrix group element."
        else:
            mat = mat.matrix();
    #Trivial Case
    if mat == mat.parent().one_element():
        return collection
    
    # Generate non existing lists
    if negative_list     == []:
        negative_list     = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list     == []:
        positive_list     = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if permutation_list  == []:
        permutation_list  = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    if row_position_list == []:
        row_position_list = genericrowoperationdict(L, negative_list=negative_list)
    
    _len = len(negative_list)
    
    _collect = collection[:]
    _out = [mat, _collect]
    print 'Transforming to upper triangular matrix:\n    Please wait!\n\n';
    for step in range(LieType(L)[1]):
        _out = rowpermutation(_out[0], L, step=step, collection=_out[1], negative_list=negative_list, positive_list=positive_list, permutation_list=permutation_list, base_ring=base_ring, sparse=sparse)
        while not is_zeroed(_out[0], step=step):
            
            _out = rowtransformation(_out[0], L, step=step,  collection=_out[1], negative_list=negative_list, positive_list=positive_list, permutation_list=permutation_list, row_position_list=row_position_list, base_ring=base_ring, sparse=sparse)
            
    _mat_out = _out[0] 
    _collect = _out[1]
    
    #print "=========================================================";
    #print "OUTPUT : After Loops";
    #print _mat_out 
    #print "=========================================================";
   
    #print "=========================================================";
    print "Eliminating elements in upper triangular matrix:\n   Please wait!\n\n"
    
    for m in row_position_list:
        _keys = m.keys();
        if len(_keys)>0:
            for k in _keys:
                _entry = _mat_out[k[1],k[0]];
                if _entry !=0:
                    _idx=m[k][0]
                    _sgn=m[k][1]
                    if abs(_sgn) != 1:
                        raise ValueError, "Entries in generators different from +-1";
                    _mat_out  = positive_list[_idx](-_sgn*_entry)*_mat_out
                    _collect.append([[_idx+_len, _sgn*_entry], positive_list [_idx] (_sgn*_entry)])

    
    
    if _mat_out!= _mat_out.parent().one_element():
        print _mat_out;
        print 'ERROR'
        return None
    print 'Decomposition succesful';
    return _collect

#===============================================================================
# getdecompositionoflist ()
#===============================================================================  

def getdecompositionoflist(lst,L, collection=[], negative_list=[], positive_list=[], permutation_list=[], row_position_list=[], base_ring=QQ, sparse=True):
    """
    See decompositionoflist, but appplies to a list of matrices and returns a list of decompositions.
    """
    
    if negative_list     == []:
        negative_list     = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list     == []:
        positive_list     = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if permutation_list  == []:
        permutation_list  = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    if row_position_list == []:
        row_position_list = genericrowoperationdict(L, negative_list=negative_list)    
    
    if type(lst)!=type([]):
        raise TypeError, "Method only works for lists. If you want to decompose a single element please use the method SymplecticDecomposition"
    
    _out = map(lambda X: getdecomposition (X, L, collection=collection, negative_list = negative_list, positive_list = positive_list, permutation_list = permutation_list,row_position_list = row_position_list, base_ring=base_ring, sparse=sparse), lst)
    
    
    
    return _out

#===============================================================================
# getreprofgroupelement ()
#===============================================================================     


def getreprofgroupelement (el, HWM, negative_list=[], positive_list=[], HWMactiongenerators=[], permutation_list=[], row_position_list=[], UEA_generators=[], base_ring=QQ, sparse=True): 
    """
    INPUT::
        el  : a Lie group element
        HWM : a highest weight module
        ...
    OUTPUT::
        a matrix which is the representation of el on the basis of HWM by multiplicatio from the left
    EXAMPLE::
        sage: L=SimpleLieAlgebra('C',2)
        sage: m=randomgroupelemenetbyLiealgebra (L); m
            [ 0  3  2  2]
            [ 1  1  0  1]
            [-2  1  2 -1]
            [-1  1  1  0]
        sage: M = HighestWeightModule(L,[0,1])
        sage: getreprofgroupelement(m, M)
            Precomputation done
            Transforming to upper triangular matrix:
            Please wait!


            Eliminating elements in upper triangular matrix:
                   Please wait!


            Decomposition succesful
               Compute product
                done
            [-3  2  4 -1  2]
            [-6  4  8 -5  6]
            [-3  2  3 -2  2]
            [-2  1  2 -1  1]
            [-1  0  2 -1  1]
    """
    
    if el==el.parent().one_element():
        _dim = HWM.Dimension()
        
        return identity_matrix(base_ring, _dim, sparse=sparse)
    
    L=HWM.LeftActingAlgebra()
    
    if negative_list     == []:
        negative_list     = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list     == []:
        positive_list     = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if permutation_list  == []:
        permutation_list  = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    if row_position_list == []:
        row_position_list = genericrowoperationdict(L, negative_list=negative_list) 
    if UEA_generators == []:
        UEA_generators = latticegeneratorsinUEA(L)
    if HWMactiongenerators == []:
        HWMactiongenerators = genericgroupgeneratorsfunctionsonHWM (HWM, base_ring=base_ring, sparse=sparse, UEA_generators=UEA_generators)    
    print "   Precomputation done"
    _dec = getdecomposition(el,L,collection=[], negative_list = negative_list, positive_list = positive_list, permutation_list = permutation_list,row_position_list = row_position_list, base_ring=base_ring, sparse=sparse)
    print "   Compute product"
    _mat = prod([HWMactiongenerators[ref[0]](ref[1]) for ref, mat in _dec])
    print "done"
    return _mat



#===============================================================================
# coordinatevector2HWMelement ()
#===============================================================================
def coordinatevector2HWMelement (v, M):
    """
    Represents a coordinate vector as a HWM element.
    """
    _bas = M.Basis()
    _v = 0*_bas[1]
    _cnt =1
    for vi in v:
        _v=_v+vi*_bas[_cnt]
        _cnt = _cnt+1
    return _v

#===============================================================================
# invariants ()
#===============================================================================

def invariants(HWM, Grp, only_generators= True, in_type=0, negative_list=[], positive_list=[], HWMactiongenerators=[], permutation_list=[], row_position_list=[], UEA_generators=[], base_ring=QQ, sparse=True):
    """
    Computes the space of the invariants of a module HWM 
    under a group Grp or a list  of elements.
    INPUT::
        M               : HWM
        Grp             : MatrixGroup, list of matrices, 
                          list of matrix group elements
        in_type         : default = 0 (Output of stabilizers is of type 0)
        only_generators : default = True, reduces the list of elements, 
                          to a list of generators
        
        Optional Parameters:
            they are computed if not known. For definition see description in         
            getreprofgroupelement.
            
        negative_list       = []
        positive_list       = []
        HWMactiongenerators = []
        permutation_list    = []
        row_position_list   = []
        UEA_generators      = []
        base_ring           = QQ
        sparse              = True
        
    OUTPUT::
        V       : Submodule of base_ring**rank(HWM) in terms of coordinate 
                  vectors corresponding to the basis of HWM
    """
    from symplecticgrp import getgenerators
    print "* Check Input and convert to list of matrices"
    V = VectorSpace(base_ring,HWM.Dimension())
    if isinstance(Grp, list):
        if Grp==[]:
            return V
        elif is_Matrix(Grp[0]):
            pass
            _G = Grp[:]
        elif is_MatrixGroupElement(Grp[0]):
            _G = [g.matrix() for g in Grp]
        else:
            raise TypeError, "Elements in the list has to be a matrix or a matrix group element"
    elif is_MatrixGroup(Grp):
        if not Grp.is_finite():
            raise NotImplementedError, "Is only implemented for finite Groups"
        else:
            _G = [g.matrix() for g in Grp]
    else:
        raise TypeError, "Input has to be either a MatrixGroup or a list of matrices or matrix group elements."

    if only_generators:
        print "* Compute List of Generators"
        #check has to be false since some of the sets are not invariant under the group.
        _G = getgenerators(_G,check=False)
        print "    --> Done"
    if in_type == 0:
        print "* Coverting matrices to the right type."
        _G = [symplectic2symplectic(g, type_from=0,type_to=1) for g in _G]
        print "   Number of generators%s"% len(_G)
        print "   -> Done"
    

    
    print "* Precompute list related to HWMs if not known"
    
    L=HWM.LeftActingAlgebra()
    if negative_list     == []:
        negative_list     = negativegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if positive_list     == []:
        positive_list     = positivegenericgroupgeneratorsfunctions(L,base_ring=base_ring)
    if permutation_list  == []:
        permutation_list  = genericpermutations(L,negative_list=negative_list,positive_list=positive_list)
    if row_position_list == []:
        row_position_list = genericrowoperationdict(L, negative_list=negative_list) 
    if UEA_generators == []:
        UEA_generators = latticegeneratorsinUEA(L)
    if HWMactiongenerators == []:
        HWMactiongenerators = genericgroupgeneratorsfunctionsonHWM (HWM, base_ring=base_ring, sparse=sparse, UEA_generators=UEA_generators)    
    
    print "    Have to handle %s elements" %len(_G)
    cnt =1
    for g0 in _G:
        print "    * Start with element %s" %cnt 
        g = getreprofgroupelement (g0, HWM, negative_list = negative_list, positive_list = positive_list, HWMactiongenerators = HWMactiongenerators, permutation_list = permutation_list, row_position_list = row_position_list, UEA_generators=[], base_ring = base_ring, sparse = sparse)
        
        id = g.parent().one_element()
        m = g-id
        mk = m.right_kernel()
        if mk == V.zero_submodule():
            return V.zero_submodule()
        V = V.intersection(mk)
        if V==V.zero_submodule():
            return V
        cnt=cnt+1
         
    print '     --> Done'
    
    return V
