
"""
symplectic - Module to decribe symplectic matrices

AUTHOR:

- JENS PUTZKA (2010-05-18): initial version

DATE:
- 2010-10-06

VERSION:
- 1.5 - 2010-10-06
- 1.0 - 2010-05-18

"""
#*****************************************************************************
#       Copyright (C) 2009 - JENS PUTZKA <putzka@mpim-bonn.mpg.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************


print "\n=========================================================";
print '|  Importing Symplectic Package  (c) 2008-11            |';
print '|                                                       |'
print '|    by Jens Putzka MPI for Mathematics Bonn            |'
print "=========================================================";

#===============================================================================
# Imports
#===============================================================================


from sage.rings.all import ZZ,QQ
from sage.matrix.all import is_Matrix, diagonal_matrix, matrix, identity_matrix
from sage.groups.all import MatrixGroup#, is_MatrixGroupElement


#===============================================================================
# PART I: Functions and Constants
#===============================================================================

#===============================================================================
# D0()
#===============================================================================

def D0 (g, symplectic_type=0):
    """
    D0 (g):
    Gives diagonalmatrix with g times 1 and g times -1 on the diagonal\n
    
    INPUT::
    g               = 1,2,...
    symplectic_type = 0 (other not complete implemented)
    
    CAVE::
    Works only if either type=0 or g = 2 
    
    EXAMPLE::
    sage: D0(2)
        [ 1  0  0  0]
        [ 0  1  0  0]
        [ 0  0 -1  0]
        [ 0  0  0 -1]
    sage: D0(3)
        [ 1  0  0  0  0  0]
        [ 0  1  0  0  0  0]
        [ 0  0  1  0  0  0]
        [ 0  0  0 -1  0  0]
        [ 0  0  0  0 -1  0]
        [ 0  0  0  0  0 -1]
    """
    if symplectic_type==0:
        _entry = diagonal_matrix([1 for i in range(g)]);
        _out   = _entry.block_sum(-_entry);
        return _out;
    elif g==2:
        _entry = diagonal_matrix([1,-1]);
        _out   = _entry.block_sum(-_entry);
        return _out;        
    else:
        raise NotImplementedError

#===============================================================================
# symplectic_J ()
#===============================================================================


def symplectic_J (g, symplectic_type=0):
    """
    symplectic_J (g):
    Gives the symplectic Matrix J of dimension 2g for some type
    
    INPUT::
    g =1,2,...
    symplectic_type=0,1
    
    EXAMPLE::
    
    sage: symplectic_J (2, symplectic_type=0)        
        [ 0  0  0  1]
        [ 0  0  1  0]
        [ 0 -1  0  0]
        [-1  0  0  0]
    sage: symplectic_J (2, symplectic_type=1)        
        [ 0  0  0  1]
        [ 0  0 -1  0]
        [ 0  1  0  0]
        [-1  0  0  0]

    """
    if symplectic_type==0:
        _out = (D0 (g, symplectic_type=0)).matrix_from_columns(range(2*g-1,-1,-1))
        return _out;
    elif symplectic_type==1:
        if g==2:
            _out = (D0 (g, symplectic_type=0)).matrix_from_columns(range(2*g-1,-1,-1))
            return _out
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError


#===============================================================================
# is_symplectic ()
#===============================================================================

def is_symplectic (matr, symplectic_type=0) :
    """
    INPUT:: a matrix  or an MatrixGroupElement mat
    OUTPUT:: Returns True if mat is symplectic and FALSE otherwise
    
    """
    if not is_Matrix(matr):
        try:
            matr=matr.matrix()
        except:
            raise TypeError, "Input has to be a matrix or a MatrixGroupElement."
    _n=matr.ncols();
    if _n not in 2*ZZ:
        if matr!=matr.parent(0):
            return False;
    elif (matr.transpose())*symplectic_J(_n/2,symplectic_type=symplectic_type)*matr==symplectic_J(_n/2,symplectic_type=symplectic_type) :
        return True;
    else :
        return False;    
    


#===============================================================================
# is_GSp ()
#===============================================================================

def is_GSp (matr, symplectic_type = 0):
    """
    INPUT::  a matrix  or an MatrixGroupElement mat
    OUTPUT:: Returns TRUE if mat is in GSp and FALSE otherwise.
    """
    if not is_Matrix(matr):
        try:
            matr=matr.matrix()
        except:
            raise TypeError, "Input has to be a matrix or a MatrixGroupElement."
    _n=matr.ncols();
    if not matr.is_invertible():
        return False;
    elif is_symplectic (matr,symplectic_type = symplectic_type) :
        return True;
    elif is_symplectic (D0(_n/2,symplectic_type=symplectic_type)*matr):
        return True;
    else :
        return False;

#===============================================================================
# symplectic2symplecticMatrix ()
#===============================================================================        

def symplectic2symplecticMatrix (g=2,type_from=0,type_to=1):
    """
    INPUT::
        g=2
        type_from = 0
        type_to   = 1
    
    Cave: Only implemented if either 
    g=2 and type_from = 0 and type_to = 1 
    or 
    type_from = type_to
    
    
    OUT::
        T = matrix such that T.transpose()*J_0*T== J_1
    
    EXAMPLE::
       sage: symplectic2symplecticMatrix()
          [1 0 0 0]
          [0 0 1 0]
          [0 1 0 0]
          [0 0 0 1]
    """
    if type_to==type_from:
        return identity_matrix(2*g)
    elif g==2:
        if type_from==0:
            if type_to==1:
                T = matrix([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]])
            else:
                raise NotImplementedError
        elif type_from==1:
            if type_to==0:
                T= matrix([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]])
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError
    return T                

#===============================================================================
# symplectic2symplectic ()
#===============================================================================        

def symplectic2symplectic (matr, type_from=0,type_to=1):
    """
    INPUT:: 
        matrix
        type_from = 0
        type_to   = 1
    OUTPUT::
        matrix with respect to a different sympletic form, i.e.
        matrix_neu = T* matrix_alt * T.inverse()
        
        where T= symplectic2symplecticMatrix(...)
    EXAMPLE::
        sage: J = symplectic_J(2)
            [ 0  0  0  1]
            [ 0  0  1  0]
            [ 0 -1  0  0]
            [-1  0  0  0]
        sage: symplectic2symplecticMatrix(J)
            [ 0  0  0  1]
            [ 0  0 -1  0]
            [ 0  1  0  0]
            [-1  0  0  0]
    """
    _g = matr.ncols()/2 
    T  = symplectic2symplecticMatrix (_g, type_from=type_from, type_to=type_to)
    return T*matr*T.inverse()

