


"""
symplectic - Module to decribe symplectic groups

AUTHOR:

- JENS PUTZKA (2010-05-18): initial version

DATE:
- 2010-10-06

VERSION:
- 2.0 - 2011-12-05
- 1.5 - 2010-10-06
- 1.0 - 2010-05-18


"""
#*****************************************************************************
#       Copyright (C) 2009 - JENS PUTZKA <putzka@mpim-bonn.mpg.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************


print "\n=========================================================";
print '|  Importing Symplectic Group Package  (c) 201          |';
print '|                                                       |'
print '|    by Jens Putzka MPI for Mathematics Bonn            |'
print "=========================================================";

#===============================================================================
# Imports
#===============================================================================


from sage.rings.all import ZZ,QQ
from sage.matrix.all import is_Matrix, diagonal_matrix, matrix, identity_matrix
from sage.groups.all import MatrixGroup#, is_MatrixGroupElement
from sage.interfaces.gap import gap
from hwmodule import genericgroupgenerators, SimpleLieAlgebra
from symplectic import symplectic_J
from sage.parallel.all import parallel

#===============================================================================
# GeneratorsSymplecticGroup ()
#===============================================================================

def GeneratorsSymplecticGroup(g, symplectic_type=0):
    """
    Returns for a g the list of generators of the group Sp_g(ZZ)
    INPUT::
        g
        symplectic_type = 0
    OUTPUT::
        list of matrices, which generates Sp_g(ZZ).
    EXAMPLE::
    
        sage: GeneratorsSymplecticGroup(2)
            [
            [1 0 1 0]  [1 0 0 1]  [1 0 0 0]  [ 0  0  0  1]
            [0 1 0 0]  [0 1 1 0]  [0 1 0 1]  [ 0  0  1  0]
            [0 0 1 0]  [0 0 1 0]  [0 0 1 0]  [ 0 -1  0  0]
            [0 0 0 1], [0 0 0 1], [0 0 0 1], [-1  0  0  0]
            ]
        sage: GeneratorsSymplecticGroup(3)
    """
    if symplectic_type==0:
        _out = [identity_matrix(2*g)+matrix(2*g,{(i,j+g):1,(j,i+g):1}) for i in range(g) for j in range (i+1)];
    
        _out.append(symplectic_J(g,symplectic_type=0));        
        
        return _out; 
    elif symplectic_type==1:
        L = SimpleLieAlgebra ("C", g)
        _gen = genericgroupgenerators(L)
        return _gen
    else:
       raise NotImplementedError
      
#===============================================================================
# SymplecticGroup ()
#===============================================================================  



def SymplecticGroup(g,symplectic_type=0):
    """
    SymplecticGroup(g)
    INPUT::
    OUTPUT::
    EXAMPLE::    
    """
    gen = GeneratorsSymplecticGroup(g,symplectic_type=symplectic_type)
    return MatrixGroup(gen);
    
#===============================================================================
# makefinitegrp ()
#===============================================================================


def makefinitegrp(grp, p=7):
    """
    INPUT::
           grp = a (finite) matrix group over ZZ
        p = 7 (default) a prime (works also for non prime numbers N)
    RETURN:: 
        a matrix group over GF(p) (or Z/NZ)
    """
    pp = ZZ(p)
    lst = [g.matrix().mod(pp)for g in grp] 
    G = MatrixGroup(lst)
    return G
    
#===============================================================================
# getgroupstructure ()
#===============================================================================    
def getgroupstructure (grp, p=7):
    G= gap(makefinitegrp(grp, p=p))
    
    return G.StructureDescription()

#===============================================================================
# intersectionofgroups ()
#===============================================================================


def intersectionofgroups (grp1,grp2):
    """
    intersectionofgroups (grp1,grp2)
    
    INPUT::
        grp1,grp2 : (finite) MatrixGroups
    OUTPUT::
        MatrixGroup which is the intesection of grp1 and grp2
    """
    out = [m.matrix() for m in grp1 if m.matrix() in [n.matrix() for n in grp2]]
    return MatrixGroup(out)

#===============================================================================
# quotientsystem ()
#===============================================================================

def quotientsystem (biggroup,smallgroup):
    """
    quotientsystem (biggroup,smallgroup)
    
    INPUT:
        biggroup a MatrixGroup
        smallgroup a MatrixGroup which is a subgroup of giggroup
    
    OUTPUT:
        List of matrices which are representatives of biggroup/smallgroup. First element in the list should be the identity_matrix
     
    """
    
    if not  is_MatrixGroup(biggroup):
        raise TypeError, "%s is  not a MatrixGroup" %biggroup;
    if not is_MatrixGroup(smallgroup):
        raise TypeError, "%s is  not a MatrixGroup" %biggroup;
    if not (biggroup.is_finite() and smallgroup.is_finite()):
        raise NotImplementedError, "Sorry at the moment only the case where both groups are finite is implemented."
    
    biglst   = [b.matrix() for b in biggroup]
    smalllst = [s.matrix() for s in smallgroup]
    outlst=[identity_matrix(biglst[1].ncols())]
    for s in smalllst:
        try :
            biglst.remove(s)
        except:
            raise ValueError, "%s has to be a subgroup of %s" %(smallgroup, biggroup)
    while biglst!=[]:
        tmp_el=biglst.pop()
        outlst.append(tmp_el)
        for s in smalllst:
            try:
                biglst.remove(tmp_el*s)
            except:
                pass
    return outlst;

#===============================================================================
# getgenerators ()
#===============================================================================
def getgenerators (G, check=False, max_cnt = 10**2):
    """
    Computes from a set of matrices G a set of generators.
        
    if check= True (False is default) it checks whether the set is
    closed under the matrix multiplication. If it is not closed it 
    returns the original list.
        
    """
        
    if check:
        try:
            _H = G[:]
            for _g1 in G:
                for _g2 in G:
                    _g12 = _g1*_g2
                    _H.remove(_g12)
                    _H.append(_g12)
        except:
            print "Set is not closed under multiplication\n therefore the output is the set itself and NOT a set of generators"
            return G
        
    __G = G[:]
        
    _id     = __G[0].parent().one()
    _gen = []
    tmp  = []
    if _id in __G:
        tmp.append(_id)
        __G.remove(_id)
    if -_id in __G:
        _gen.append(-_id)
        tmp.append(-_id)
        __G.remove(-_id)
        
           
    while __G!=[]:
        #take 1st element
        _g = __G.pop()
        _gen.append(_g)
        tmp.append(_g)
        _gprod =_g
        if __G==[]:
            return _gen
        # remove powers of th element
        cnt = 0
        while _gprod != _id:
            if cnt > max_cnt:
                print "Computation of generators failed."
                raise ValueError, "Maximal recursion depth exceeded. Pobable there is an element of infinite order"
            cnt=cnt+1
            _gprod = _gprod * _g
            if _gprod in __G:
                __G.remove(_gprod)
            if _gprod not in tmp:
                tmp.append(_gprod)    
            if __G==[]:
                return _gen
             
        for _h1 in tmp:
            for _h2 in tmp:
                _h12 = _h1*_h2
                if _h12!=_id:
                    if _h12 not in tmp:
                        tmp.append(_h12)
                        if _h12 in __G:
                            __G.remove(_h12)
                            if __G==[]:
                                return _gen
            
    
    return _gen